from typing import Type, List, Callable

import pytest  # type: ignore

from remerkleable.basic import uint, uint8, uint16, uint32, uint64, uint128, uint256, \
    OperationNotSupported

uint_types = [uint8, uint16, uint32, uint64, uint128, uint256]

uint_valid_cases: List[Callable[[Type[uint]], int]] = [
    lambda t: 0,
    lambda t: 1,
    lambda t: 3,
    lambda t: 53,
    lambda t: (1 << (8 * t.type_byte_length())) // 10,
    lambda t: (1 << (8 * t.type_byte_length())) // 2,
    lambda t: (1 << (8 * t.type_byte_length()))-1,
]

bi_operations = ['add', 'sub', 'floordiv', 'mul', 'mod', 'and', 'xor', 'or']
reverse_bi_operations = [f'r{op}' for op in bi_operations]


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
@pytest.mark.parametrize("b", uint_valid_cases)
@pytest.mark.parametrize("op", bi_operations + reverse_bi_operations)
@pytest.mark.parametrize("b_unsigned", [True, False])
def test_uint_arithmetic(typ, a, b, op, b_unsigned):
    a_v = a(typ)
    b_v = b(typ)
    uint_a = typ(a_v)
    uint_b = typ(b_v)
    f = getattr(a_v, f'__{op}__')
    regular_err = None
    int_v = -1
    try:
        int_v = f(b_v)
    except Exception as e:
        regular_err = e

    err = None
    uint_v = None
    try:
        uint_f = getattr(uint_a, f'__{op}__')
        uint_v = uint_f(uint_b if b_unsigned else b_v)
    except Exception as e:
        err = e

    # E.g. divide by zero, modulo errors, etc. If it doesn't work for a python int, it shouldn't work for a uint as well
    if err is not None and (err == regular_err or isinstance(err, type(regular_err))):
        return

    if int_v < 0 or int_v >= (1 << (8 * typ.type_byte_length())):
        assert isinstance(err, ValueError)
    else:
        assert int(uint_v) == int_v
        assert uint_v == typ(int_v)
        assert isinstance(uint_v, typ)


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("v", [-256, -255, -3, -1])
def test_uint_lower_bound(typ, v):
    try:
        typ(v)
        raise Exception('expected value error')
    except ValueError:
        pass


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("vf", [
    lambda t: (1 << (8 * t.type_byte_length())),
    lambda t: (1 << (8 * t.type_byte_length())) + 1,
    lambda t: (1 << (16 * t.type_byte_length())),
])
def test_uint_upper_bound(typ, vf):
    try:
        int_v = vf(typ)
        typ(int_v)
        raise Exception('expected value error')
    except ValueError:
        pass


shift_cases: List[Callable[[Type[uint]], int]] = [
    lambda t: 0,
    lambda t: 1,
    lambda t: 2,
    lambda t: 8,
    lambda t: (8 * t.type_byte_length()) // 10,
    lambda t: (8 * t.type_byte_length()) // 2,
    lambda t: (8 * t.type_byte_length()) - 1,
    lambda t: (8 * t.type_byte_length()),
    lambda t: (8 * t.type_byte_length()) + 1,
    lambda t: (8 * t.type_byte_length()) * 2,
]

shift_operations = ['lshift', 'rshift']


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
@pytest.mark.parametrize("b", shift_cases)
@pytest.mark.parametrize("op", shift_operations)
@pytest.mark.parametrize("b_unsigned", [True, False])
def test_uint_shifts(typ, a, b, op, b_unsigned):
    a_v = a(typ)
    b_v = b(typ)
    uint_a = typ(a_v)
    uint_b = typ(b_v)
    f = getattr(a_v, f'__{op}__')
    int_v = f(b_v)
    uint_f = getattr(uint_a, f'__{op}__')
    uint_v = uint_f(uint_b if b_unsigned else b_v)
    mask = (1 << (typ.type_byte_length() << 3)) - 1
    assert mask.to_bytes(length=typ.type_byte_length(), byteorder='little').hex() == 'ff' * typ.type_byte_length()
    assert (int_v & mask) == int(uint_v)


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
@pytest.mark.parametrize("b", shift_cases)
@pytest.mark.parametrize("op", shift_operations)
@pytest.mark.parametrize("a_unsigned", [True, False])
def test_uint_rshifts(typ, a, b, op, a_unsigned):
    a_v = a(typ)
    b_v = b(typ)
    uint_a = typ(a_v)
    uint_b = typ(b_v)
    f = getattr(a_v, f'__{op}__')
    int_v = f(b_v)
    uint_f = getattr(uint_b, f'__r{op}__')
    try:
        uint_v = uint_f(uint_a if a_unsigned else a_v)
        # Only allow reverse shifts if the left operand is a uint, otherwise we can't know the resulting size.
        assert a_unsigned
        mask = (1 << (typ.type_byte_length() << 3)) - 1
        assert mask.to_bytes(length=typ.type_byte_length(), byteorder='little').hex() == 'ff' * typ.type_byte_length()
        assert (int_v & mask) == int(uint_v)
    except ValueError:
        assert not a_unsigned


def test_mixed_uint_lshifts():
    assert uint64(0xff) << uint8(8*7) == uint64(0xff00_0000_0000_0000)
    assert uint16(0xff) << uint8(8*7) == uint16(0)
    assert uint256(1) << uint8(255) == uint256(1 << 255)
    assert uint256(1) << uint16(255) == uint256(1 << 255)
    assert uint256(1) << uint16(256) == uint256(0)
    assert uint8(42) >> uint64(3) == uint8(42 >> 3)


def test_mixed_uint_rshifts():
    assert uint64(0xff00_0000_0000_0000) >> uint8(8*7) == uint64(0xff)
    assert uint16(0xff) >> uint8(8) == uint16(0)
    assert uint16(0xff) >> uint8(6) == uint16(3)
    assert uint256(1 << 255) >> uint8(255) == uint256(1)
    assert uint256(1 << 255) >> uint16(255) == uint256(1)
    assert uint8(42) << uint64(3) == uint8((42 << 3) & 0xff)


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
@pytest.mark.parametrize("b_v", [0, 1, 2, 3, 5])
@pytest.mark.parametrize("rev", [True, False])
@pytest.mark.parametrize("b_unsigned", [True, False])
def test_uint_pow(typ, a, b_v, rev, b_unsigned):
    a_v = a(typ)
    uint_a = typ(a_v)
    uint_b = typ(b_v)

    regular_err = None
    int_v = -1
    try:
        int_v = a_v**b_v
    except Exception as e:
        regular_err = e

    err = None
    uint_v = None
    try:
        uint_v = uint_a**(uint_b if b_unsigned else b_v)
    except Exception as e:
        err = e

    # If python ints can't handle it either, then it's ok
    if err is not None and (err == regular_err or isinstance(err, type(regular_err))):
        return

    if int_v < 0 or int_v >= (1 << (8 * typ.type_byte_length())):
        assert isinstance(err, ValueError)
    else:
        assert int(uint_v) == int_v
        assert uint_v == typ(int_v)
        assert isinstance(uint_v, typ)


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
def test_uint_negative(typ, a):
    a_v = a(typ)
    uint_a = typ(a_v)
    try:
        x = -uint_a
        raise Exception(f"expected OperationNotSupported exception, but got result {x}")
    except OperationNotSupported:
        pass


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
@pytest.mark.parametrize("op", ['truediv', 'rtruediv'])
def test_uint_bi_op_unsupported(typ, a, op):
    a_v = a(typ)
    uint_a = typ(a_v)
    uint_f = getattr(uint_a, f'__{op}__')
    try:
        x = uint_f(42)
        raise Exception(f"expected OperationNotSupported exception, but got result {x}")
    except OperationNotSupported:
        pass


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
def test_uint_identity(typ, a):
    a_v = a(typ)
    uint_a = typ(a_v)
    assert a_v == abs(uint_a)
    assert a_v == +uint_a


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("a", uint_valid_cases)
def test_uint_invert(typ, a):
    a_v = a(typ)
    uint_a = typ(a_v)
    inverted_a = ~uint_a
    mask = (1 << (typ.type_byte_length() << 3)) - 1
    assert uint_a | inverted_a == mask
    assert inverted_a != uint_a
    bitlen = uint_a.type_byte_length() * 8
    bits = [(1 << i) & uint_a != 0 for i in range(bitlen)]
    inverted_bits = [(1 << i) & inverted_a != 0 for i in range(bitlen)]
    assert all(bits[i] != inverted_bits[i] for i in range(bitlen))


@pytest.mark.parametrize("typ", uint_types)
@pytest.mark.parametrize("count", [0, 1, 42])
def test_uint_list_mul(typ, count):
    uint_count = typ(count)
    data = [None] * uint_count
    assert len(data) == count



from typing import Any, TypeVar, Type
from remerkleable.core import BasicView, View, ObjType, ObjParseException

V = TypeVar('V', bound=View)


# Not returning "NotImplemented" like regular operators,
# it's completely invalid, do not let the interpreter resort to the other operation hand.
class OperationNotSupported(Exception):
    pass


BoolV = TypeVar('BoolV', bound="boolean")


class boolean(int, BasicView):

    def encode_bytes(self) -> bytes:
        return b"\x01" if self else b"\x00"

    def __new__(cls, value: int):  # int value, but can be any subclass of int (bool, Bit, Bool, etc...)
        if value < 0 or value > 1:
            raise ValueError(f"value {value} out of bounds for bit")
        return super().__new__(cls, value)  # type: ignore

    def __add__(self, other):
        raise OperationNotSupported(f"cannot add bool ({self} + {other})")

    def __sub__(self, other):
        raise OperationNotSupported(f"cannot sub bool ({self} - {other})")

    def __mul__(self, other):
        raise OperationNotSupported(f"cannot mul bool ({self} * {other})")

    def __floordiv__(self, other):  # Better known as "//"
        raise OperationNotSupported(f"cannot floordiv bool ({self} // {other})")

    def __truediv__(self, other):
        raise OperationNotSupported(f"cannot truediv bool ({self} / {other})")

    def __bool__(self):
        return self > 0

    @classmethod
    def coerce_view(cls: Type[BoolV], v: Any) -> BoolV:
        return cls(v)

    @classmethod
    def type_byte_length(cls) -> int:
        return 1

    @classmethod
    def decode_bytes(cls: Type[BoolV], bytez: bytes) -> BoolV:
        return cls(bytez != b"\x00")

    @classmethod
    def from_obj(cls: Type[BoolV], obj: ObjType) -> BoolV:
        if not isinstance(obj, bool):
            raise ObjParseException(f"obj '{obj}' is not a bool")
        return cls(obj)

    def to_obj(self) -> ObjType:
        return bool(self)

    @classmethod
    def type_repr(cls) -> str:
        return "boolean"


T = TypeVar('T', bound="uint")
W = TypeVar('W', bound=int)


class uint(int, BasicView):
    def __new__(cls, value: int):
        if value < 0:
            raise ValueError(f"unsigned type {cls} must not be negative")
        byte_len = cls.type_byte_length()
        if value.bit_length() > (byte_len << 3):
            raise ValueError(f"value out of bounds for {cls}")
        return super().__new__(cls, value)  # type: ignore

    def __add__(self: T, other: int) -> T:
        return self.__class__(super().__add__(self.__class__.coerce_view(other)))

    def __radd__(self: T, other: int) -> T:
        return self.__add__(other)

    def __sub__(self: T, other: int) -> T:
        return self.__class__(super().__sub__(self.__class__.coerce_view(other)))

    def __rsub__(self: T, other: int) -> T:
        return self.__class__(self.__class__.coerce_view(other).__sub__(self))

    def __mul__(self, other):
        if not isinstance(other, int):
            return super().__mul__(other)
        return self.__class__(super().__mul__(self.__class__.coerce_view(other)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mod__(self: T, other: int) -> T:
        return self.__class__(super().__mod__(self.__class__.coerce_view(other)))

    def __rmod__(self: T, other: int) -> T:
        return self.__class__(self.__class__.coerce_view(other).__mod__(self))

    def __floordiv__(self: T, other: int) -> T:  # Better known as "//"
        return self.__class__(super().__floordiv__(self.__class__.coerce_view(other)))

    def __rfloordiv__(self: T, other: int) -> T:
        return self.__class__(self.__class__.coerce_view(other).__floordiv__(self))

    def __truediv__(self, other):
        raise OperationNotSupported(f"non-integer division '{self} / {other}' "
                                    f"is not valid for {self.__class__.type_repr()} left hand type")

    def __rtruediv__(self, other):
        raise OperationNotSupported(f"non-integer division '{other} / {self}' "
                                    f"is not valid for {self.__class__.type_repr()} right hand type")

    def __pow__(self: T, other: int, modulo=None) -> T:
        return self.__class__(super().__pow__(other, modulo))  # TODO: stricter argument checks?

    def __rpow__(self: T, other, modulo=None) -> T:
        return self.__class__(super().__rpow__(other, modulo))  # TODO: see __pow__

    def __lshift__(self: T, other: int) -> T:
        """Left bitshift clips bits at uint boundary"""
        mask = (1 << (self.type_byte_length() << 3)) - 1
        return self.__class__(super().__lshift__(int(other)) & mask)

    def __rlshift__(self, other: W) -> W:
        if not isinstance(other, uint):
            raise ValueError(f"{other} << {self} through __rlshift__ is not supported, "
                             f"left operand {other} must be a uint type with __lshift__")
        return other.__lshift__(self)  # type: ignore

    def __rshift__(self: T, other: int) -> T:
        return self.__class__(super().__rshift__(int(other)))

    def __rrshift__(self, other: W) -> W:
        if not isinstance(other, uint):
            raise ValueError(f"{other} >> {self} through __rrshift__ is not supported, "
                             f"left operand {other} must be a uint type with __rshift__")
        return other.__rshift__(self)  # type: ignore

    def __and__(self: T, other: int) -> T:
        return self.__class__(super().__and__(self.__class__.coerce_view(other)))

    def __rand__(self: T, other: int) -> T:
        return self.__and__(other)

    def __xor__(self: T, other: int) -> T:
        return self.__class__(super().__xor__(self.__class__.coerce_view(other)))

    def __rxor__(self: T, other: int) -> T:
        return self.__xor__(other)

    def __or__(self: T, other: int) -> T:
        return self.__class__(super().__or__(self.__class__.coerce_view(other)))

    def __ror__(self: T, other: int) -> T:
        return self.__or__(other)

    def __neg__(self):
        raise OperationNotSupported(f"Cannot make uint type negative! If intentional, cast to signed int first.")

    def __invert__(self: T) -> T:
        mask = (1 << (self.type_byte_length() << 3)) - 1
        return self.__xor__(mask)

    def __pos__(self: T) -> T:
        return self

    def __abs__(self: T) -> T:
        return self

    # __coerce__ is avoided to utilize explicit type hinting and special case the edge-cases for unsigned int safety

    @classmethod
    def coerce_view(cls: Type[T], v: Any) -> T:
        if isinstance(v, uint) and cls.type_byte_length() != v.__class__.type_byte_length():
            raise ValueError("value must have equal byte length to coerce it")
        if isinstance(v, bytes):
            return cls.decode_bytes(v)
        return cls(v)

    @classmethod
    def decode_bytes(cls: Type[T], bytez: bytes) -> T:
        return cls(int.from_bytes(bytez, byteorder='little'))

    def encode_bytes(self) -> bytes:
        return self.to_bytes(length=self.__class__.type_byte_length(), byteorder='little')

    @classmethod
    def from_obj(cls: Type[T], obj: ObjType) -> T:
        if not isinstance(obj, (int, str)):
            raise ObjParseException(f"obj '{obj}' is not an int or str")
        if isinstance(obj, str):
            if obj.startswith('0x'):
                return cls.decode_bytes(bytes.fromhex(obj[2:]))
            obj = int(obj)
        return cls(obj)

    def to_obj(self) -> ObjType:
        return int(self)

    @classmethod
    def type_repr(cls) -> str:
        return f"uint{cls.type_byte_length()*8}"


class uint8(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 1


class uint16(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 2


class uint32(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 4


class uint64(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 8

    # JSON encoder should be able to handle uint64, converting it to a string if necessary.
    # no "to_obj" here.


class uint128(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 16

    def to_obj(self) -> ObjType:
        return "0x" + self.encode_bytes().hex()


class uint256(uint):
    @classmethod
    def type_byte_length(cls) -> int:
        return 32

    def to_obj(self) -> ObjType:
        return "0x" + self.encode_bytes().hex()


class bit(boolean):
    pass


class byte(uint8):
    pass



from typing import List as PyList, Optional, Type, Sequence
from remerkleable.core import BasicView, View
from remerkleable.tree import Node, Root, RootNode, ZERO_ROOT


class BitfieldIter(object):
    """Iterates a subtree by traversing it with a stack (thus readonly), returning bits"""

    anchor: Node
    depth: int
    i: int
    j: int
    rootIndex: int
    length: int
    currentRoot: Root
    stack: PyList[Optional[Node]]

    def __init__(self, anchor: Node, depth: int, length: int):
        self.anchor = anchor
        self.depth = depth
        self.length = length
        self.i = 0
        self.j = 0
        self.rootIndex = 0
        self.currentRoot = ZERO_ROOT
        self.stack = [None] * depth
        limit = (1 << depth) << 8
        if limit < length:
            raise Exception(f"cannot handle iterate {length} bits in subtree of depth {depth} deep (limit {limit} bits)")

    def __iter__(self):
        self.i = 0
        self.j = 0
        self.rootIndex = 0
        self.currentRoot = ZERO_ROOT
        self.stack = [None] * self.depth
        return self

    def __next__(self):
        # done yet?
        if self.i >= self.length:
            raise StopIteration

        # in the middle of a node currently? finish that first
        if self.j > 0:
            elByte = self.currentRoot[self.j >> 3]
            elem = ((elByte >> (self.j & 7)) & 1) == 1
            self.j += 1
            if self.j > 0xff:
                self.j = 0
            self.i += 1
            return elem

        stackIndex: int
        if self.rootIndex != 0:
            # XOR current index with previous index
            # Result: highest bit matches amount we have to backtrack up the stack
            s = self.rootIndex ^ (self.rootIndex - 1)
            stackIndex = self.depth
            while s != 0:
                s >>= 1
                stackIndex -= 1

            # then move to the right from that upper previously remembered left-hand node
            node = self.stack[stackIndex]
            node = node.get_right()
            stackIndex += 1
        else:
            node = self.anchor
            stackIndex = 0

        # and move down left into this new subtree
        for x in range(stackIndex, self.depth):
            # remember left-hand nodes, we may revisit them
            self.stack[x] = node
            node = node.get_left()

        # Get leaf node as a root
        if not node.is_leaf():
            raise Exception(f"expected leaf node {self.rootIndex} to be a Root type")

        # remember the root, we need it for more bits
        self.currentRoot = node.root

        # get the first bit
        el = (self.currentRoot[0] & 1 == 1)
        # indicate that we have done one bit, and need to read more
        self.j = 1
        # And that the next root will be 1 after
        self.rootIndex += 1
        # And we progress the general element counter
        self.i += 1

        # Return the actual element
        return el


class PackedIter(object):
    """Iterates a subtree by traversing it with a stack (thus readonly), returning the unpacked elements"""

    anchor: Node
    depth: int
    i: int
    j: int
    rootIndex: int
    length: int
    per_node: int
    currentRoot: RootNode
    elem_type: Type[BasicView]
    stack: PyList[Optional[Node]]

    def __init__(self, anchor: Node, depth: int, length: int, elem_type: Type[BasicView]):
        self.anchor = anchor
        self.depth = depth
        self.length = length
        self.i = 0
        self.rootIndex = 0
        self.currentRoot = RootNode(ZERO_ROOT)
        self.stack = [None] * depth
        self.elem_type = elem_type

        self.per_node = 32 // elem_type.type_byte_length()
        self.j = self.per_node

        limit = (1 << depth) * self.per_node

        if limit < length:
            raise Exception(f"cannot handle iterate length {length} bottom subviews ({self.per_node} per node) "
                            f"in subtree of depth {depth} deep (limit {limit} subviews)")

    def __iter__(self):
        self.i = 0
        self.j = self.per_node
        self.rootIndex = 0
        self.currentRoot = RootNode(ZERO_ROOT)
        self.stack = [None] * self.depth
        return self

    def __next__(self):
        # done yet?
        if self.i >= self.length:
            raise StopIteration

        # in the middle of a node currently? finish that first
        if self.j < self.per_node:
            elem = self.elem_type.basic_view_from_backing(self.currentRoot, self.j)
            self.j += 1
            self.i += 1
            return elem

        stackIndex: int
        if self.rootIndex != 0:
            # XOR current index with previous index
            # Result: highest bit matches amount we have to backtrack up the stack
            s = self.rootIndex ^ (self.rootIndex - 1)
            stackIndex = self.depth
            while s != 0:
                s >>= 1
                stackIndex -= 1

            # then move to the right from that upper previously remembered left-hand node
            node = self.stack[stackIndex]
            node = node.get_right()
            stackIndex += 1
        else:
            node = self.anchor
            stackIndex = 0

        # and move down left into this new subtree
        for x in range(stackIndex, self.depth):
            # remember left-hand nodes, we may revisit them
            self.stack[x] = node
            node = node.get_left()

        # Get leaf node as a root
        if not node.is_leaf():
            raise Exception(f"expected leaf node {self.rootIndex} to be a Root type")

        # remember the root, we need it for more elements
        self.currentRoot = node

        # get the first element
        el = self.elem_type.basic_view_from_backing(node, 0)
        # indicate that we have done one packed element, and need to read more
        self.j = 1
        # And that the next root will be 1 after
        self.rootIndex += 1
        # And we progress the general element counter
        self.i += 1

        # Return the actual element
        return el


class NodeIter(object):
    """Iterates a subtree by traversing it with a stack (thus readonly), returning the bottom nodes"""

    anchor: Node
    depth: int
    i: int
    length: int
    per_node: int
    stack: PyList[Optional[Node]]

    def __init__(self, anchor: Node, depth: int, length: int):
        self.anchor = anchor
        self.depth = depth
        self.length = length
        self.i = 0
        self.currentRoot = RootNode(ZERO_ROOT)
        self.stack = [None] * depth

        limit = 1 << depth
        if limit < length:
            raise Exception(f"cannot handle iterate length {length} bottom nodes "
                            f"in subtree of depth {depth} deep (limit {limit} nodes)")

    def __iter__(self):
        self.i = 0
        self.stack = [None] * self.depth
        return self

    def __next__(self):
        # done yet?
        if self.i >= self.length:
            raise StopIteration

        stackIndex: int
        if self.i != 0:
            # XOR current index with previous index
            # Result: highest bit matches amount we have to backtrack up the stack
            s = self.i ^ (self.i - 1)
            stackIndex = self.depth
            while s != 0:
                s >>= 1
                stackIndex -= 1

            # then move to the right from that upper previously remembered left-hand node
            node = self.stack[stackIndex]
            node = node.get_right()
            stackIndex += 1
        else:
            node = self.anchor
            stackIndex = 0

        # and move down left into this new subtree
        for x in range(stackIndex, self.depth):
            # remember left-hand nodes, we may revisit them
            self.stack[x] = node
            node = node.get_left()

        # We progress the general element counter
        self.i += 1

        return node


class ComplexElemIter(NodeIter):
    """Iterates a subtree by traversing it with a stack (thus readonly), reusing a view to return elements"""

    reused_view: View

    def __init__(self, anchor: Node, depth: int, length: int, elem_type: Type[View]):
        super().__init__(anchor, depth, length)
        self.reused_view = elem_type.default(None)

    def __next__(self):
        node = super().__next__()
        self.reused_view.set_backing(node)
        return self.reused_view


class ComplexFreshElemIter(NodeIter):
    """Iterates a subtree by traversing it with a stack (thus readonly), not reusing a view to return elements"""

    elem_type: Type[View]

    def __init__(self, anchor: Node, depth: int, length: int, elem_type: Type[View]):
        super().__init__(anchor, depth, length)
        self.elem_type = elem_type

    def __next__(self):
        node = super().__next__()
        return self.elem_type.view_from_backing(node, None)


class ContainerElemIter(NodeIter):
    """Iterates a subtree by traversing it with a stack (thus readonly), not reusing a view to return elements"""

    elem_types: Sequence[Type[View]]

    def __init__(self, anchor: Node, depth: int, elem_types: Sequence[Type[View]]):
        super().__init__(anchor, depth, len(elem_types))
        self.elem_types = elem_types

    def __next__(self):
        i = self.i
        node = super().__next__()
        return self.elem_types[i].view_from_backing(node, None)



from typing import Callable, NewType, List, Optional, Protocol, TypeVar, Iterable, Iterator, Tuple
from hashlib import sha256


# Get the depth required for a given element count
# (in out): (0 0), (1 0), (2 1), (3 2), (4 2), (5 3), (6 3), (7 3), (8 3), (9 4)
def get_depth(elem_count: int) -> int:
    if elem_count <= 1:
        return 0
    return (elem_count - 1).bit_length()


Gindex = NewType("Gindex", int)

ROOT_GINDEX = Gindex(1)
LEFT_GINDEX = Gindex(2)
RIGHT_GINDEX = Gindex(3)


def to_gindex(index: int, depth: int) -> Gindex:
    anchor = 1 << depth
    if index >= anchor:
        raise Exception("index %d too large for depth %d" % (index, depth))
    return Gindex(anchor | index)


def get_anchor_gindex(gindex: Gindex) -> Gindex:
    return Gindex(1 << (gindex.bit_length() - 1))


def gindex_bit_iter(gindex: Gindex) -> Tuple[Iterator[bool], int]:
    if gindex < 1:
        raise Exception(f"invalid gindex: {gindex}")
    bit_len = gindex.bit_length()

    def iter_bits():
        if bit_len <= 1:
            return
        shift_v = 1 << (bit_len - 2)
        while shift_v != 0:
            yield (gindex & shift_v) != 0
            shift_v >>= 1
    return iter_bits(), bit_len - 1


def concat_gindices(steps: Iterable[Gindex]) -> Gindex:
    out = 1
    for step in steps:
        step_bit_len = step.bit_length() - 1
        out <<= step_bit_len
        out |= step ^ (1 << step_bit_len)
    return Gindex(out)


Root = NewType("Root", bytes)

MerkleFn = Callable[[Root, Root], Root]

ZERO_ROOT: Root = Root(b'\x00' * 32)


def merkle_hash(left: Root, right: Root) -> Root:
    return Root(sha256(left + right).digest())


Link = Callable[["Node"], "Node"]
SummaryLink = Callable[[], "Node"]


class Node(Protocol):

    def get_left(self) -> "Node":
        raise NavigationError

    def get_right(self) -> "Node":
        raise NavigationError

    def getter(self, target: Gindex) -> "Node":
        if target < 1:
            raise NavigationError
        if target == 1:
            return self
        node = self
        bit_iter, _ = gindex_bit_iter(target)
        for bit in bit_iter:
            if bit:
                node = node.get_right()
            else:
                node = node.get_left()
        return node

    def is_leaf(self) -> bool:
        return False

    def rebind_left(self, v: "Node") -> "Node":
        raise NavigationError

    def rebind_right(self, v: "Node") -> "Node":
        raise NavigationError

    def setter(self, target: Gindex, expand: bool = False) -> Link:
        raise NavigationError

    def summarize_into(self, target: Gindex) -> SummaryLink:
        setter = self.setter(target)
        node = self.getter(target)
        return lambda: setter(RootNode(node.merkle_root()))

    @property
    def root(self) -> Root:
        return self.merkle_root()

    def merkle_root(self) -> Root:
        raise


# hashes of hashes of zeroes etc.
zero_hashes: List[Root] = [ZERO_ROOT]

for i in range(100):
    zero_hashes.append(merkle_hash(zero_hashes[i], zero_hashes[i]))


def zero_node(depth: int) -> "RootNode":
    return RootNode(zero_hashes[depth])


def identity(v: Node) -> Node:
    return v


def compose(inner: Link, outer: Link) -> Link:
    return lambda v: outer(inner(v))


class NavigationError(RuntimeError):
    pass


V = TypeVar('V', bound=Node)


class RebindableNode(Node):
    def combine(self, left: Node, right: Node) -> Node:
        return PairNode(left, right)

    def rebind_left(self, v: Node) -> Node:
        return self.combine(v, self.get_right())

    def rebind_right(self, v: Node) -> Node:
        return self.combine(self.get_left(), v)

    def setter(self, target: Gindex, expand: bool = False) -> Link:
        if target < 1:
            raise NavigationError
        if target == 1:
            return identity
        if target == 2:
            return self.rebind_left
        if target == 3:
            return self.rebind_right
        bit_iter, depth = gindex_bit_iter(target)
        first = bit_iter.__next__()
        link = self.rebind_right if first else self.rebind_left
        prev_bit = first
        node: Node = self
        for bit in bit_iter:
            if prev_bit:
                node = node.get_right()
            else:
                node = node.get_left()
            depth -= 1
            if node.is_leaf():
                if not expand:
                    raise NavigationError
                child = zero_node(depth - 1)
                node = self.combine(child, child)
            # Ignored typing, since the rebind methods are `Callable[[Arg(Node, 'v')], Node]` (mypy specific)
            # instead of `Callable[[Node], Node]` (aka `Link`)
            if bit:
                link = compose(node.rebind_right, link)  # type: ignore
            else:
                link = compose(node.rebind_left, link)  # type: ignore
            prev_bit = bit
        return link


class PairNode(RebindableNode, Node):
    """An optimized, with lazily-computed root, a node that references two child nodes."""

    __slots__ = 'left', 'right', '_root'

    left: Node
    right: Node
    _root: Optional[Root]

    def __init__(self, left: Node, right: Node):
        self.left = left
        self.right = right
        self._root = None

    def get_left(self) -> "Node":
        return self.left

    def get_right(self) -> "Node":
        return self.right

    def is_leaf(self) -> bool:
        return False

    def merkle_root(self) -> Root:
        if self._root is not None:
            return self._root
        self._root = merkle_hash(self.left.merkle_root(), self.right.merkle_root())
        return self._root

    def __repr__(self) -> str:
        return f"H({self.left}, {self.right})"


def subtree_fill_to_depth(bottom: Node, depth: int) -> Node:
    node = bottom
    while depth > 0:
        node = PairNode(node, node)
        depth -= 1
    return node


def subtree_fill_to_length(bottom: Node, depth: int, length: int) -> Node:
    if length == 0:
        return zero_node(depth)
    if length > (1 << depth):
        raise Exception("too many nodes")
    if length == (1 << depth):
        return subtree_fill_to_depth(bottom, depth)
    if depth == 0:
        if length == 1:
            return bottom
        else:
            raise NavigationError
    if depth == 1:
        return PairNode(bottom, bottom if length > 1 else zero_node(0))
    else:
        anchor = 1 << depth
        pivot = anchor >> 1
        if length <= pivot:
            return PairNode(subtree_fill_to_length(bottom, depth - 1, length), zero_node(depth))
        else:
            return PairNode(
                subtree_fill_to_depth(bottom, depth-1),
                subtree_fill_to_length(bottom, depth-1, length - pivot)
            )


def subtree_fill_to_contents(nodes: List[Node], depth: int) -> Node:
    if len(nodes) == 0:
        return zero_node(depth)
    if len(nodes) > (1 << depth):
        raise Exception("too many nodes")
    if depth == 0:
        if len(nodes) == 1:
            return nodes[0]
        else:
            raise NavigationError
    if depth == 1:
        return PairNode(nodes[0], nodes[1] if len(nodes) > 1 else zero_node(0))
    else:
        anchor = 1 << depth
        pivot = anchor >> 1
        if len(nodes) <= pivot:
            return PairNode(subtree_fill_to_contents(nodes, depth - 1), zero_node(depth - 1))
        else:
            return PairNode(
                subtree_fill_to_contents(nodes[:pivot], depth-1),
                subtree_fill_to_contents(nodes[pivot:], depth-1)
            )


class RootNode(Node):
    """An optimized root-holding node. To check if a Node functions as node without children,
     use node.is_leaf(), since there may be more classes implementing non-child node behavior."""

    __slots__ = '_root'

    _root: Root

    def __init__(self, root: Root):
        self._root = root

    def getter(self, target: Gindex) -> Node:
        if target != 1:
            raise NavigationError
        return self

    def is_leaf(self) -> bool:
        return True

    def setter(self, target: Gindex, expand: bool = False) -> Link:
        if target < 1:
            raise NavigationError
        if target == 1:
            return identity
        if expand:
            child = zero_node(target.bit_length() - 2)
            return PairNode(child, child).setter(target, expand=True)
        else:
            raise NavigationError

    @property
    def root(self) -> Root:
        # Override to directly provide the root instead of through merkle_root()
        return self._root

    def merkle_root(self) -> Root:
        return self._root

    def __repr__(self):
        return f"0x{self._root.hex()}"


def leaf_iter(node: Node) -> Iterator[Node]:
    """Iterate ove the leaf nodes of the given node. Left-to-right order."""
    if node.is_leaf():
        yield node
        return
    yield from leaf_iter(node.get_left())
    yield from leaf_iter(node.get_right())


def get_diff(a: Node, b: Node) -> Iterator[Tuple[Node, Node]]:
    """Iterate over the changes of b, not common with a. Left-to-right order.
     Returns (a,b) tuples that can't be diffed deeper."""
    if a.root != b.root:
        if a.is_leaf() or b.is_leaf():
            yield a, b
        else:
            yield from get_diff(a.get_left(), b.get_left())
            yield from get_diff(a.get_right(), b.get_right())



from typing import Protocol, Optional
from remerkleable.tree import Node, Root, RebindableNode, NavigationError


class VirtualSource(Protocol):
    def get_left(self, key: Root) -> Node:
        ...

    def get_right(self, key: Root) -> Node:
        ...

    def is_leaf(self, key: Root) -> bool:
        ...


class VirtualNode(RebindableNode, Node):
    """A node that instead of lazily computing the root, lazily fetches the left and right child based on the root."""

    _root: Root
    _src: VirtualSource
    _is_leaf: Optional[bool] = None
    _left: Optional[Node] = None
    _right: Optional[Node] = None

    def __init__(self, root: Root, src: VirtualSource):
        self._root = root
        self._src = src

    def get_left(self) -> Node:
        if self._left is None:
            if self._is_leaf is None or self._is_leaf is False:
                self._left = self._src.get_left(self._root)
            else:
                raise NavigationError("Virtual node is a leaf node, cannot get left child node.")
        return self._left

    def get_right(self) -> Node:
        if self._right is None:
            if self._is_leaf is None or self._is_leaf is False:
                self._right = self._src.get_right(self._root)
            else:
                raise NavigationError("Virtual node is a leaf node, cannot get right child node.")
        return self._right

    def is_leaf(self) -> bool:
        if self._is_leaf is None:
            self._is_leaf = self._src.is_leaf(self._root)
        return self._is_leaf

    @property
    def root(self) -> Root:
        return self._root

    def merkle_root(self) -> Root:
        return self._root

    def __repr__(self):
        return f"0x{self._root.hex()}"



from typing import NamedTuple, cast, List as PyList, Dict, Any, BinaryIO, Optional,\
    TypeVar, Type, Protocol, runtime_checkable
from types import GeneratorType
from textwrap import indent
from collections.abc import Sequence as ColSequence
from itertools import chain
import io
from remerkleable.core import View, BasicView, OFFSET_BYTE_LENGTH, ViewHook, ObjType, ObjParseException
from remerkleable.basic import uint256, uint8, uint32
from remerkleable.tree import Node, subtree_fill_to_length, subtree_fill_to_contents,\
    zero_node, Gindex, PairNode, to_gindex, NavigationError, get_depth
from remerkleable.subtree import SubtreeView
from remerkleable.readonly_iters import PackedIter, ComplexElemIter, ComplexFreshElemIter, ContainerElemIter

V = TypeVar('V', bound=View)


def decode_offset(stream: BinaryIO) -> uint32:
    return cast(uint32, uint32.deserialize(stream, OFFSET_BYTE_LENGTH))


def encode_offset(stream: BinaryIO, offset: int):
    return uint32(offset).serialize(stream)


class ComplexView(SubtreeView):
    def encode_bytes(self) -> bytes:
        stream = io.BytesIO()
        self.serialize(stream)
        stream.seek(0)
        return stream.read()

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        stream = io.BytesIO()
        stream.write(bytez)
        stream.seek(0)
        return cls.deserialize(stream, len(bytez))


M = TypeVar('M', bound="MonoSubtreeView")


class MonoSubtreeView(ColSequence, ComplexView):

    def length(self) -> int:
        raise NotImplementedError

    @classmethod
    def coerce_view(cls: Type[M], v: Any) -> M:
        return cls(*v)

    @classmethod
    def element_cls(cls) -> Type[View]:
        raise NotImplementedError

    @classmethod
    def item_elem_cls(cls, i: int) -> Type[View]:
        return cls.element_cls()

    @classmethod
    def to_chunk_length(cls, elems_length: int) -> int:
        if cls.is_packed():
            elem_type: Type[View] = cls.element_cls()
            if issubclass(elem_type, BasicView):
                elems_per_chunk = 32 // elem_type.type_byte_length()
                return (elems_length + elems_per_chunk - 1) // elems_per_chunk
            else:
                raise Exception("cannot append a packed element that is not a basic type")
        else:
            return elems_length

    @classmethod
    def views_into_chunks(cls, views: PyList[View]) -> PyList[Node]:
        if cls.is_packed():
            elem_type: Type[View] = cls.element_cls()
            if issubclass(elem_type, BasicView):
                # cast the list as a whole, checking each element takes too long.
                return elem_type.pack_views(cast(PyList[BasicView], views))
            else:
                raise Exception("cannot append a packed element that is not a basic type")
        else:
            return [v.get_backing() for v in views]

    @classmethod
    def is_valid_count(cls, count: int) -> bool:
        raise NotImplementedError

    def __iter__(self):
        return iter(self.get(i) for i in range(self.length()))

    def readonly_iter(self):
        tree_depth = self.tree_depth()
        length = self.length()
        backing = self.get_backing()

        elem_type: Type[View] = self.element_cls()

        if self.is_packed():
            return PackedIter(backing, tree_depth, length, cast(Type[BasicView], elem_type))
        else:
            if issubclass(elem_type, bytes):  # is the element type the raw-bytes? Then not re-use views.
                return ComplexFreshElemIter(backing, tree_depth, length, cast(Type[View], elem_type))
            else:
                return ComplexElemIter(backing, tree_depth, length, elem_type)

    @classmethod
    def deserialize(cls: Type[M], stream: BinaryIO, scope: int) -> M:
        elem_cls = cls.element_cls()
        if elem_cls.is_fixed_byte_length():
            elem_byte_length = elem_cls.type_byte_length()
            if scope % elem_byte_length != 0:
                raise Exception(f"scope {scope} does not match element byte length {elem_byte_length} multiple")
            count = scope // elem_byte_length
            if not cls.is_valid_count(count):
                raise Exception(f"count {count} is invalid")
            return cls(elem_cls.deserialize(stream, elem_byte_length) for _ in range(count))  # type: ignore
        else:
            if scope == 0:
                if not cls.is_valid_count(0):
                    raise Exception("scope cannot be 0, count must not be 0")
                return cls()
            first_offset = decode_offset(stream)
            if first_offset > scope:
                raise Exception(f"first offset is too big: {first_offset}, scope: {scope}")
            if first_offset % OFFSET_BYTE_LENGTH != 0:
                raise Exception(f"first offset {first_offset} is not a multiple of offset length {OFFSET_BYTE_LENGTH}")
            count = first_offset // OFFSET_BYTE_LENGTH
            if not cls.is_valid_count(count):
                raise Exception(f"count {count} is invalid")
            # count - 1: we already have the first offset
            offsets = [first_offset] + [decode_offset(stream) for _ in range(count - 1)] + [uint32(scope)]
            elem_min, elem_max = elem_cls.min_byte_length(), elem_cls.max_byte_length()
            elems = []
            for i in range(count):
                start, end = offsets[i], offsets[i+1]
                if end < start:
                    raise Exception(f"offsets[{i}] value {start} is invalid, next offset is {end}")
                elem_size = end - start
                if not (elem_min <= elem_size <= elem_max):
                    raise Exception(f"offset[{i}] value {start} is invalid, next offset is {end},"
                                    f" implied size is {elem_size}, size bounds: [{elem_min}, {elem_max}]")
                elems.append(elem_cls.deserialize(stream, elem_size))
            return cls(*elems)  # type: ignore

    def serialize(self, stream: BinaryIO) -> int:
        elem_cls = self.__class__.element_cls()
        if issubclass(elem_cls, uint8):
            out = bytes(iter(self))
            stream.write(out)
            return len(out)
        if elem_cls.is_fixed_byte_length():
            for v in self.readonly_iter():
                v.serialize(stream)
            return elem_cls.type_byte_length() * self.length()
        else:
            temp_dyn_stream = io.BytesIO()
            offset = OFFSET_BYTE_LENGTH * self.length()  # the offsets are part of the fixed-size-bytes prologue
            for v in self:
                encode_offset(stream, offset)
                offset += cast(View, v).serialize(temp_dyn_stream)
            temp_dyn_stream.seek(0)
            stream.write(temp_dyn_stream.read(offset))
            return offset

    @classmethod
    def from_obj(cls: Type[M], obj: ObjType) -> M:
        if not isinstance(obj, (list, tuple)):
            raise ObjParseException(f"obj '{obj}' is not a list or tuple")
        elem_cls = cls.element_cls()
        return cls(elem_cls.from_obj(el) for el in obj)  # type: ignore

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        if key < 0:
            raise KeyError
        return cls.element_cls()

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        if key < 0:
            raise KeyError

        if cls.is_packed():
            elems_per_chunk = 32 // cls.element_cls().type_byte_length()
            chunk_i = key // elems_per_chunk
        else:
            chunk_i = key

        return to_gindex(chunk_i, cls.tree_depth())

    def navigate_view(self, key: Any) -> View:
        return self.__getitem__(key)

    def __len__(self):
        return self.length()

    def __add__(self, other):
        if issubclass(self.element_cls(), uint8):
            return bytes(self) + bytes(other)
        else:
            return list(chain(self, other))

    def __getitem__(self, k):
        if isinstance(k, slice):
            start = 0 if k.start is None else k.start
            end = self.length() if k.stop is None else k.stop
            return [self.get(i) for i in range(start, end)]
        else:
            return self.get(k)

    def __setitem__(self, k, v):
        if type(k) == slice:
            i = 0 if k.start is None else k.start
            end = self.length() if k.stop is None else k.stop
            for item in v:
                self.set(i, item)
                i += 1
            if i != end:
                raise Exception("failed to do full slice-set, not enough values")
        else:
            self.set(k, v)

    def _repr_sequence(self):
        length: int
        try:
            length = self.length()
        except NavigationError:
            return f"{self.type_repr()}( *summary root, no length known* )"
        vals: Dict[int, View] = {}
        partial = False
        for i in range(length):
            try:
                vals[i] = self.get(i)
            except NavigationError:
                partial = True
                continue
        basic_elems = isinstance(self.element_cls(), BasicView)
        shortened = length > (64 if basic_elems else 8)
        summary_length = (10 if basic_elems else 3)
        seperator = ', ' if basic_elems else ',\n'
        contents = seperator.join(f"... {length - (summary_length * 2)} omitted ..."
                                  if (shortened and i == summary_length)
                                  else (f"{i}: {repr(v)}" if partial else repr(v))
                                  for i, v in vals.items()
                                  if (not shortened) or i <= summary_length or i >= length - summary_length)
        if '\n' in contents:
            contents = '\n' + indent(contents, '  ') + '\n'
        if partial:
            return f"{self.type_repr()}~partial~<<len={length}>>({contents})"
        else:
            return f"{self.type_repr()}<<len={length}>>({contents})"


class List(MonoSubtreeView):
    def __new__(cls, *args, backing: Optional[Node] = None, hook: Optional[ViewHook] = None, **kwargs):
        if backing is not None:
            if len(args) != 0:
                raise Exception("cannot have both a backing and elements to init List")
            return super().__new__(cls, backing=backing, hook=hook, **kwargs)

        elem_cls = cls.element_cls()
        vals = list(args)
        if len(vals) == 1:
            val = vals[0]
            if isinstance(val, (GeneratorType, list, tuple)):
                vals = list(val)
            if issubclass(elem_cls, uint8):
                if isinstance(val, bytes):
                    vals = list(val)
                if isinstance(val, str):
                    if val[:2] == '0x':
                        val = val[2:]
                    vals = list(bytes.fromhex(val))
        if len(vals) > 0:
            limit = cls.limit()
            if len(vals) > limit:
                raise Exception(f"too many list inputs: {len(vals)}, limit is: {limit}")
            input_views = []
            for el in vals:
                if isinstance(el, View):
                    input_views.append(el)
                else:
                    input_views.append(elem_cls.coerce_view(el))
            input_nodes = cls.views_into_chunks(input_views)
            contents = subtree_fill_to_contents(input_nodes, cls.contents_depth())
            backing = PairNode(contents, uint256(len(input_views)).get_backing())
        return super().__new__(cls, backing=backing, hook=hook, **kwargs)

    def __class_getitem__(cls, params) -> Type["List"]:
        (element_type, limit) = params
        contents_depth = 0
        packed = False
        if isinstance(element_type, BasicView):
            elems_per_chunk = 32 // element_type.type_byte_length()
            contents_depth = get_depth((limit + elems_per_chunk - 1) // elems_per_chunk)
            packed = True
        else:
            contents_depth = get_depth(limit)

        class SpecialListView(List):
            @classmethod
            def is_packed(cls) -> bool:
                return packed

            @classmethod
            def contents_depth(cls) -> int:
                return contents_depth

            @classmethod
            def element_cls(cls) -> Type[View]:
                return element_type

            @classmethod
            def limit(cls) -> int:
                return limit

        SpecialListView.__name__ = SpecialListView.type_repr()
        return SpecialListView

    def length(self) -> int:
        ll_node = super().get_backing().get_right()
        ll = cast(uint256, uint256.view_from_backing(node=ll_node, hook=None))
        return int(ll)

    def value_byte_length(self) -> int:
        elem_cls = self.__class__.element_cls()
        if elem_cls.is_fixed_byte_length():
            return elem_cls.type_byte_length() * self.length()
        else:
            return sum(OFFSET_BYTE_LENGTH + cast(View, el).value_byte_length() for el in iter(self))

    def append(self, v: View):
        ll = self.length()
        if ll >= self.__class__.limit():
            raise Exception("list is maximum capacity, cannot append")
        i = ll
        elem_type: Type[View] = self.__class__.element_cls()
        if not isinstance(v, elem_type):
            v = elem_type.coerce_view(v)
        target: Gindex
        if self.__class__.is_packed():
            next_backing = self.get_backing()
            if isinstance(v, BasicView):
                elems_per_chunk = 32 // elem_type.type_byte_length()
                chunk_i = i // elems_per_chunk
                target = to_gindex(chunk_i, self.__class__.tree_depth())
                chunk: Node
                if i % elems_per_chunk == 0:
                    set_last = next_backing.setter(target, expand=True)
                    chunk = zero_node(0)
                else:
                    set_last = next_backing.setter(target)
                    chunk = next_backing.getter(target)
                chunk = v.backing_from_base(chunk, i % elems_per_chunk)
                next_backing = set_last(chunk)
            else:
                raise Exception("cannot append a packed element that is not a basic type")
        else:
            target = to_gindex(i, self.__class__.tree_depth())
            set_last = self.get_backing().setter(target, expand=True)
            next_backing = set_last(v.get_backing())

        set_length = next_backing.rebind_right
        new_length = uint256(ll + 1).get_backing()
        next_backing = set_length(new_length)
        self.set_backing(next_backing)

    def pop(self):
        ll = self.length()
        if ll == 0:
            raise Exception("list is empty, cannot pop")
        i = ll - 1
        target: Gindex
        can_summarize: bool
        if self.__class__.is_packed():
            next_backing = self.get_backing()
            elem_type: Type[View] = self.__class__.element_cls()
            if issubclass(elem_type, BasicView):
                elems_per_chunk = 32 // elem_type.type_byte_length()
                chunk_i = i // elems_per_chunk
                target = to_gindex(chunk_i, self.__class__.tree_depth())
                if i % elems_per_chunk == 0:
                    chunk = zero_node(0)
                else:
                    chunk = next_backing.getter(target)
                set_last = next_backing.setter(target)
                chunk = elem_type.default(None).backing_from_base(chunk, i % elems_per_chunk)
                next_backing = set_last(chunk)

                can_summarize = (target & 1) == 0 and i % elems_per_chunk == 0
            else:
                raise Exception("cannot pop a packed element that is not a basic type")
        else:
            target = to_gindex(i, self.__class__.tree_depth())
            set_last = self.get_backing().setter(target)
            next_backing = set_last(zero_node(0))
            can_summarize = (target & 1) == 0

        # if possible, summarize
        if can_summarize:
            # summarize to the highest node possible.
            # I.e. the resulting target must be a right-hand, unless it's the only content node.
            while (target & 1) == 0 and target != 0b10:
                target >>= 1
            summary_fn = next_backing.summarize_into(target)
            next_backing = summary_fn()

        set_length = next_backing.rebind_right
        new_length = uint256(ll - 1).get_backing()
        next_backing = set_length(new_length)
        self.set_backing(next_backing)

    def get(self, i: int) -> View:
        if i < 0 or i >= self.length():
            raise IndexError
        return super().get(i)

    def set(self, i: int, v: View) -> None:
        if i < 0 or i >= self.length():
            raise IndexError
        super().set(i, v)

    def __repr__(self):
        return self._repr_sequence()

    @classmethod
    def type_repr(cls) -> str:
        return f"List[{cls.element_cls().__name__}, {cls.limit()}]"

    @classmethod
    def is_packed(cls) -> bool:
        raise NotImplementedError

    @classmethod
    def contents_depth(cls) -> int:
        raise NotImplementedError

    @classmethod
    def tree_depth(cls) -> int:
        return cls.contents_depth() + 1  # 1 extra for length mix-in

    @classmethod
    def item_elem_cls(cls, i: int) -> Type[View]:
        return cls.element_cls()

    @classmethod
    def limit(cls) -> int:
        raise NotImplementedError

    @classmethod
    def is_valid_count(cls, count: int) -> bool:
        return 0 <= count <= cls.limit()

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        if key >= cls.limit():
            raise KeyError
        return super().navigate_type(key)

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        if key >= cls.limit():
            raise KeyError
        return super().key_to_static_gindex(key)

    @classmethod
    def default_node(cls) -> Node:
        return PairNode(zero_node(cls.contents_depth()), zero_node(0))  # mix-in 0 as list length

    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        return False

    @classmethod
    def min_byte_length(cls) -> int:
        return 0

    @classmethod
    def max_byte_length(cls) -> int:
        elem_cls = cls.element_cls()
        bytes_per_elem = elem_cls.max_byte_length()
        if not elem_cls.is_fixed_byte_length():
            bytes_per_elem += OFFSET_BYTE_LENGTH
        return bytes_per_elem * cls.limit()

    def to_obj(self) -> ObjType:
        return list(el.to_obj() for el in self.readonly_iter())


class Vector(MonoSubtreeView):
    def __new__(cls, *args, backing: Optional[Node] = None, hook: Optional[ViewHook] = None, **kwargs):
        if backing is not None:
            if len(args) != 0:
                raise Exception("cannot have both a backing and elements to init Vector")
            return super().__new__(cls, backing=backing, hook=hook, **kwargs)

        elem_cls = cls.element_cls()
        vals = list(args)
        if len(vals) == 1:
            val = vals[0]
            if isinstance(val, (GeneratorType, list, tuple)):
                vals = list(val)
            if issubclass(elem_cls, uint8):
                if isinstance(val, bytes):
                    vals = list(val)
                if isinstance(val, str):
                    if val[:2] == '0x':
                        val = val[2:]
                    vals = list(bytes.fromhex(val))
        if len(vals) > 0:
            vector_length = cls.vector_length()
            if len(vals) != vector_length:
                raise Exception(f"invalid inputs length: {len(vals)}, vector length is: {vector_length}")
            input_views = []
            for el in vals:
                if isinstance(el, View):
                    input_views.append(el)
                else:
                    input_views.append(elem_cls.coerce_view(el))
            input_nodes = cls.views_into_chunks(input_views)
            backing = subtree_fill_to_contents(input_nodes, cls.tree_depth())
        return super().__new__(cls, backing=backing, hook=hook, **kwargs)

    def __class_getitem__(cls, params) -> Type["Vector"]:
        (element_view_cls, length) = params
        if length <= 0:
            raise Exception(f"Invalid vector length: {length}")

        tree_depth = 0
        packed = False
        if isinstance(element_view_cls, BasicView):
            elems_per_chunk = 32 // element_view_cls.type_byte_length()
            tree_depth = get_depth((length + elems_per_chunk - 1) // elems_per_chunk)
            packed = True
        else:
            tree_depth = get_depth(length)

        class SpecialVectorView(Vector):
            @classmethod
            def is_packed(cls) -> bool:
                return packed

            @classmethod
            def tree_depth(cls) -> int:
                return tree_depth

            @classmethod
            def element_cls(cls) -> Type[View]:
                return element_view_cls

            @classmethod
            def vector_length(cls) -> int:
                return length

        out_typ = SpecialVectorView

        # for fixed-size vectors, pre-compute the size.
        if element_view_cls.is_fixed_byte_length():
            byte_length = element_view_cls.type_byte_length() * length

            class FixedSpecialVectorView(SpecialVectorView):
                @classmethod
                def type_byte_length(cls) -> int:
                    return byte_length

                @classmethod
                def min_byte_length(cls) -> int:
                    return byte_length

                @classmethod
                def max_byte_length(cls) -> int:
                    return byte_length

            out_typ = FixedSpecialVectorView

        out_typ.__name__ = out_typ.type_repr()
        return out_typ

    def get(self, i: int) -> View:
        if i < 0 or i >= self.__class__.vector_length():
            raise IndexError
        return super().get(i)

    def set(self, i: int, v: View) -> None:
        if i < 0 or i >= self.__class__.vector_length():
            raise IndexError
        super().set(i, v)

    def length(self) -> int:
        return self.__class__.vector_length()

    def value_byte_length(self) -> int:
        if self.__class__.is_fixed_byte_length():
            return self.__class__.type_byte_length()
        else:
            return sum(OFFSET_BYTE_LENGTH + cast(View, el).value_byte_length() for el in iter(self))

    def __repr__(self):
        return self._repr_sequence()

    @classmethod
    def type_repr(cls) -> str:
        return f"Vector[{cls.element_cls().__name__}, {cls.vector_length()}]"

    @classmethod
    def vector_length(cls) -> int:
        raise NotImplementedError

    @classmethod
    def is_valid_count(cls, count: int) -> bool:
        return count == cls.vector_length()

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        if key >= cls.vector_length():
            raise KeyError
        return super().navigate_type(key)

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        if key >= cls.vector_length():
            raise KeyError
        return super().key_to_static_gindex(key)

    @classmethod
    def default_node(cls) -> Node:
        elem_type: Type[View] = cls.element_cls()
        length = cls.to_chunk_length(cls.vector_length())
        elem: Node
        if cls.is_packed():
            elem = zero_node(0)
        else:
            elem = elem_type.default_node()
        return subtree_fill_to_length(elem, cls.tree_depth(), length)

    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        return cls.element_cls().is_fixed_byte_length()  # only if the element type is fixed byte length.

    @classmethod
    def min_byte_length(cls) -> int:
        elem_cls = cls.element_cls()
        bytes_per_elem = elem_cls.min_byte_length()
        if not elem_cls.is_fixed_byte_length():
            bytes_per_elem += OFFSET_BYTE_LENGTH
        return bytes_per_elem * cls.vector_length()

    @classmethod
    def max_byte_length(cls) -> int:
        elem_cls = cls.element_cls()
        bytes_per_elem = elem_cls.max_byte_length()
        if not elem_cls.is_fixed_byte_length():
            bytes_per_elem += OFFSET_BYTE_LENGTH
        return bytes_per_elem * cls.vector_length()

    def to_obj(self) -> ObjType:
        return tuple(el.to_obj() for el in self.readonly_iter())


Fields = Dict[str, Type[View]]


class FieldOffset(NamedTuple):
    key: str
    typ: Type[View]
    offset: int


@runtime_checkable
class _ContainerLike(Protocol):
    @classmethod
    def fields(cls) -> Fields:
        ...


CV = TypeVar('CV', bound="Container")


class _ContainerBase(ComplexView):
    def __new__(cls, *args, backing: Optional[Node] = None, hook: Optional[ViewHook] = None,
                append_nodes: Optional[PyList[Node]] = None, **kwargs):
        if backing is not None:
            if len(args) != 0 or append_nodes is not None:
                raise Exception("cannot have both a backing and elements to init fields")
            return super().__new__(cls, backing=backing, hook=hook, **kwargs)
        if append_nodes is None:
            raise Exception("cannot init container without fields")

        backing = subtree_fill_to_contents(append_nodes, cls.tree_depth())
        out = super().__new__(cls, backing=backing, hook=hook)
        return out

    @classmethod
    def fields(cls) -> Fields:  # base condition for the subclasses deriving the fields
        return {}


class Container(_ContainerBase):
    _field_indices: Dict[str, int]

    def __new__(cls, *args, backing: Optional[Node] = None, hook: Optional[ViewHook] = None,
                append_nodes: Optional[PyList[Node]] = None, **kwargs):
        if backing is not None:
            if len(args) != 0 or append_nodes is not None:
                raise Exception("cannot have both a backing and elements to init fields")
            return super().__new__(cls, backing=backing, hook=hook, **kwargs)

        input_nodes = []
        for fkey, ftyp in cls.fields().items():
            fnode: Node
            if fkey in kwargs:
                finput = kwargs.pop(fkey)
                if isinstance(finput, View):
                    fnode = finput.get_backing()
                else:
                    fnode = ftyp.coerce_view(finput).get_backing()
            else:
                fnode = ftyp.default_node()
            input_nodes.append(fnode)
        # if this is the base of some container subclass, add the subclass nodes to the backing we are building.
        if append_nodes is not None:
            input_nodes.extend(append_nodes)

        # check if any keys are remaining to catch unrecognized keys
        if len(kwargs) > 0:
            raise AttributeError(f'The field names [{"".join(kwargs.keys())}] are not defined in {cls}')

        out = super().__new__(cls, hook=hook, append_nodes=input_nodes)
        return out

    def __init_subclass__(cls, *args, **kwargs):
        super().__init_subclass__(*args, **kwargs)
        cls._field_indices = {fkey: i for i, fkey in enumerate(cls.fields())}
        if len(cls._field_indices) == 0:
            raise Exception(f"Container {cls.__name__} must have at least one field!")

    @classmethod
    def coerce_view(cls: Type[CV], v: Any) -> CV:
        return cls(**{fkey: getattr(v, fkey) for fkey in cls.fields().keys()})

    @classmethod
    def fields(cls) -> Fields:
        fields = {}
        for b in cls.__bases__:
            for k, v in b.fields().items():
                fields[k] = v
        for k, v in cls.__annotations__.items():
            if k[0] != '_':
                fields[k] = v  # if the key exists, overwrite it. Otherwise it extends the (ordered) dict.
        return fields

    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        return all(f.is_fixed_byte_length() for f in cls.fields().values())

    @classmethod
    def type_byte_length(cls) -> int:
        if cls.is_fixed_byte_length():
            return cls.min_byte_length()
        else:
            raise Exception("dynamic length container does not have a fixed byte length")

    @classmethod
    def min_byte_length(cls) -> int:
        total = 0
        for ftyp in cls.fields().values():
            if not ftyp.is_fixed_byte_length():
                total += OFFSET_BYTE_LENGTH
            total += ftyp.min_byte_length()
        return total

    @classmethod
    def max_byte_length(cls) -> int:
        total = 0
        for ftyp in cls.fields().values():
            if not ftyp.is_fixed_byte_length():
                total += OFFSET_BYTE_LENGTH
            total += ftyp.max_byte_length()
        return total

    @classmethod
    def is_packed(cls) -> bool:
        return False

    @classmethod
    def tree_depth(cls) -> int:
        return get_depth(len(cls.fields()))

    @classmethod
    def item_elem_cls(cls, i: int) -> Type[View]:
        return list(cls.fields().values())[i]

    @classmethod
    def default_node(cls) -> Node:
        return subtree_fill_to_contents([field.default_node() for field in cls.fields().values()], cls.tree_depth())

    def value_byte_length(self) -> int:
        if self.__class__.is_fixed_byte_length():
            return self.__class__.type_byte_length()
        else:
            total = 0
            fields = self.fields()
            for fkey, ftyp in fields.items():
                if ftyp.is_fixed_byte_length():
                    total += ftyp.type_byte_length()
                else:
                    total += OFFSET_BYTE_LENGTH
                    total += cast(View, getattr(self, fkey)).value_byte_length()
            return total

    def __getattr__(self, item):
        if item[0] == '_':
            return super().__getattribute__(item)
        else:
            try:
                i = self.__class__._field_indices[item]
            except KeyError:
                raise AttributeError(f"unknown attribute {item}")
            return super().get(i)

    def __setattr__(self, key, value):
        if key[0] == '_':
            super().__setattr__(key, value)
        else:
            try:
                i = self.__class__._field_indices[key]
            except KeyError:
                raise AttributeError(f"unknown attribute {key}")
            super().set(i, value)

    def _get_field_val_repr(self, fkey: str, ftype: Type[View]) -> str:
        field_start = '  ' + fkey + ': ' + ftype.__name__ + ' = '
        try:
            field_repr = repr(getattr(self, fkey))
            if '\n' in field_repr:  # if multiline, indent it, but starting from the value.
                i = field_repr.index('\n')
                field_repr = field_repr[:i+1] + indent(field_repr[i+1:], ' ' * len(field_start))
            return field_start + field_repr
        except NavigationError:
            return f"{field_start} *omitted from partial*"

    def __repr__(self):
        return f"{self.__class__.__name__}(Container)\n" + '\n'.join(
            indent(self._get_field_val_repr(fkey, ftype), '  ')
            for fkey, ftype in self.__class__.fields().items())

    @classmethod
    def type_repr(cls) -> str:
        return f"{cls.__name__}(Container)\n" + '\n'.join(
            ('  ' + fkey + ': ' + ftype.__name__) for fkey, ftype in cls.fields().items())

    def __iter__(self):
        tree_depth = self.tree_depth()
        backing = self.get_backing()
        return ContainerElemIter(backing, tree_depth, list(self.__class__.fields().values()))

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        stream = io.BytesIO()
        stream.write(bytez)
        stream.seek(0)
        return cls.deserialize(stream, len(bytez))

    @classmethod
    def deserialize(cls: Type[CV], stream: BinaryIO, scope: int) -> CV:
        fields = cls.fields()
        field_values: Dict[str, View]
        if cls.is_fixed_byte_length():
            field_values = {fkey: ftyp.deserialize(stream, ftyp.type_byte_length()) for fkey, ftyp in fields.items()}
        else:
            field_values = {}
            dyn_fields: PyList[FieldOffset] = []
            fixed_size = 0
            for fkey, ftyp in fields.items():
                if ftyp.is_fixed_byte_length():
                    fsize = ftyp.type_byte_length()
                    field_values[fkey] = ftyp.deserialize(stream, fsize)
                    fixed_size += fsize
                else:
                    dyn_fields.append(FieldOffset(key=fkey, typ=ftyp, offset=int(decode_offset(stream))))
                    fixed_size += OFFSET_BYTE_LENGTH
            if len(dyn_fields) > 0:
                if dyn_fields[0].offset < fixed_size:
                    raise Exception(f"first offset is smaller than expected fixed size")
                for i, (fkey, ftyp, foffset) in enumerate(dyn_fields):
                    next_offset = dyn_fields[i + 1].offset if i + 1 < len(dyn_fields) else scope
                    if foffset > next_offset:
                        raise Exception(f"offset {i} is invalid: {foffset} larger than next offset {next_offset}")
                    fsize = next_offset - foffset
                    f_min_size, f_max_size = ftyp.min_byte_length(), ftyp.max_byte_length()
                    if not (f_min_size <= fsize <= f_max_size):
                        raise Exception(f"offset {i} is invalid, size out of bounds: {foffset}, next {next_offset},"
                                        f" implied size: {fsize}, size bounds: [{f_min_size}, {f_max_size}]")
                    field_values[fkey] = ftyp.deserialize(stream, fsize)
        return cls(**field_values)  # type: ignore

    def serialize(self, stream: BinaryIO) -> int:
        fields = self.__class__.fields()
        is_fixed_size = self.is_fixed_byte_length()
        temp_dyn_stream: BinaryIO
        written = sum(map((lambda x: x.type_byte_length() if x.is_fixed_byte_length() else OFFSET_BYTE_LENGTH),
                          fields.values()))
        if not is_fixed_size:
            temp_dyn_stream = io.BytesIO()
        for fkey, ftyp in fields.items():
            v: View = getattr(self, fkey)
            if ftyp.is_fixed_byte_length():
                v.serialize(stream)
            else:
                encode_offset(stream, written)
                written += v.serialize(temp_dyn_stream)  # type: ignore
        if not is_fixed_size:
            temp_dyn_stream.seek(0)
            stream.write(temp_dyn_stream.read(written))
        return written

    @classmethod
    def from_obj(cls: Type[CV], obj: ObjType) -> CV:
        if not isinstance(obj, dict):
            raise ObjParseException(f"obj '{obj}' is not a dict")
        fields = cls.fields()
        for k in obj.keys():
            if k not in fields:
                raise ObjParseException(f"obj '{obj}' has unknown key {k}")
        return cls(**{k: fields[k].from_obj(v) for k, v in obj.items()})  # type: ignore

    def to_obj(self) -> ObjType:
        return {f_k: f_v.to_obj() for f_k, f_v in zip(self.__class__.fields().keys(), self.__iter__())}

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        fields = cls.fields()
        try:
            field_index = list(fields.keys()).index(key)
        except ValueError:  # list.index raises ValueError if the element (a key here) is missing
            raise KeyError
        return to_gindex(field_index, cls.tree_depth())

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        return cls.fields()[key]

    def navigate_view(self, key: Any) -> View:
        return self.__getattr__(key)



from typing import Optional, Any, TypeVar, Type, BinaryIO
from types import GeneratorType
from remerkleable.tree import Node, RootNode, Root, subtree_fill_to_contents, get_depth, to_gindex, \
    subtree_fill_to_length, Gindex, PairNode
from remerkleable.core import View, ViewHook, zero_node, FixedByteLengthViewHelper, pack_bytes_to_chunks, ObjType, \
    ObjParseException
from remerkleable.basic import byte, uint256


RV = TypeVar('RV', bound="RawBytesView")


class RawBytesView(bytes, View):
    def __new__(cls, *args, **kwargs):
        if len(args) == 0:
            return super().__new__(cls, cls.default_bytes(), **kwargs)
        elif len(args) == 1:
            args = args[0]
            if isinstance(args, (GeneratorType, list, tuple)):
                data = bytes(args)
            elif isinstance(args, bytes):
                data = args
            elif isinstance(args, str):
                if args[:2] == '0x':
                    args = args[2:]
                data = list(bytes.fromhex(args))
            else:
                data = bytes(args)
            return super().__new__(cls, data, **kwargs)
        else:
            return super().__new__(cls, bytes(args), **kwargs)

    @classmethod
    def default_bytes(cls) -> bytes:
        raise NotImplementedError

    @classmethod
    def coerce_view(cls: Type[RV], v: Any) -> RV:
        return cls(v)

    @classmethod
    def tree_depth(cls) -> int:
        raise NotImplementedError

    def set_backing(self, value):
        raise Exception("cannot change the backing of a raw-bytes-like view, init a new view instead")

    def __repr__(self):
        return "0x" + self.hex()

    def __str__(self):
        return "0x" + self.hex()

    @classmethod
    def decode_bytes(cls: Type[RV], bytez: bytes) -> RV:
        return cls(bytez)

    def encode_bytes(self) -> bytes:
        return self

    @classmethod
    def from_obj(cls: Type[RV], obj: ObjType) -> RV:
        if not isinstance(obj, (list, tuple, str, bytes)):
            raise ObjParseException(f"obj '{obj}' is not a list, tuple, str or bytes")
        return cls(obj)

    def to_obj(self) -> ObjType:
        return '0x' + self.encode_bytes().hex()

    def navigate_view(self, key: Any) -> View:
        return byte(self.__getitem__(key))


BV = TypeVar('BV', bound="ByteVector")


class ByteVector(RawBytesView, FixedByteLengthViewHelper, View):
    def __new__(cls, *args, **kwargs):
        byte_len = cls.vector_length()
        out = super().__new__(cls, *args, **kwargs)
        if len(out) != byte_len:
            raise Exception(f"incorrect byte length: {len(out)}, expected {byte_len}")
        return out

    def __class_getitem__(cls, length) -> Type["ByteVector"]:
        chunk_count = (length + 31) // 32
        tree_depth = get_depth(chunk_count)

        class SpecialByteVectorView(ByteVector):
            @classmethod
            def default_node(cls) -> Node:
                return subtree_fill_to_length(zero_node(0), tree_depth, chunk_count)

            @classmethod
            def tree_depth(cls) -> int:
                return tree_depth

            @classmethod
            def type_byte_length(cls) -> int:
                return length

        return SpecialByteVectorView

    @classmethod
    def vector_length(cls):
        return cls.type_byte_length()

    @classmethod
    def default_bytes(cls) -> bytes:
        return b"\x00" * cls.vector_length()

    @classmethod
    def type_repr(cls) -> str:
        return f"ByteVector[{cls.vector_length()}]"

    @classmethod
    def view_from_backing(cls: Type[BV], node: Node, hook: Optional[ViewHook] = None) -> BV:
        depth = cls.tree_depth()
        byte_len = cls.vector_length()
        if depth == 0:
            return cls.decode_bytes(node.merkle_root()[:byte_len])
        else:
            chunk_count = (byte_len + 31) // 32
            chunks = [node.getter(to_gindex(i, depth)) for i in range(chunk_count)]
            bytez = b"".join(ch.merkle_root() for ch in chunks)[:byte_len]
            return cls.decode_bytes(bytez)

    def get_backing(self) -> Node:
        if len(self) == 32:  # super common case, optimize for it
            return RootNode(Root(self))
        elif len(self) < 32:
            return RootNode(Root(self + b"\x00" * (32 - len(self))))
        else:
            return subtree_fill_to_contents(pack_bytes_to_chunks(self), self.__class__.tree_depth())

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        if key < 0 or key > cls.vector_length():
            raise KeyError
        return byte

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        depth = cls.tree_depth()
        byte_len = cls.vector_length()
        if key < 0 or key >= byte_len:
            raise KeyError
        chunk_i = key // 32
        return to_gindex(chunk_i, depth)


# Define common special Byte vector view types, these are bytes-like:
# raw representation instead of backed by a binary tree. Inheriting Python "bytes"
Bytes1 = ByteVector[1]  # type: ignore
Bytes4 = ByteVector[4]  # type: ignore
Bytes8 = ByteVector[8]  # type: ignore
Bytes32 = ByteVector[32]  # type: ignore
Bytes48 = ByteVector[48]  # type: ignore
Bytes96 = ByteVector[96]  # type: ignore


BL = TypeVar('BL', bound="ByteList")


class ByteList(RawBytesView, FixedByteLengthViewHelper, View):

    def __new__(cls, *args, **kwargs):
        byte_limit = cls.limit()
        out = super().__new__(cls, *args, **kwargs)
        if len(out) > byte_limit:
            raise Exception(f"incorrect byte length: {len(out)}, cannot be more than limit {byte_limit}")
        return out

    def __class_getitem__(cls, limit) -> Type["ByteList"]:
        chunk_count = (limit + 31) // 32
        contents_depth = get_depth(chunk_count)

        class SpecialByteListView(ByteList):
            @classmethod
            def contents_depth(cls) -> int:
                return contents_depth

            @classmethod
            def limit(cls) -> int:
                return limit

        return SpecialByteListView

    @classmethod
    def limit(cls) -> int:
        raise NotImplementedError

    @classmethod
    def default_bytes(cls) -> bytes:
        return b""

    @classmethod
    def type_repr(cls) -> str:
        return f"ByteList[{cls.limit()}]"

    @classmethod
    def view_from_backing(cls: Type[BL], node: Node, hook: Optional[ViewHook] = None) -> BL:
        contents_depth = cls.contents_depth()
        contents_node = node.get_left()
        length = uint256.view_from_backing(node.get_right())
        if length > cls.limit():
            raise Exception("ByteList backing declared length exceeds limit")
        if contents_depth == 0:
            return cls.decode_bytes(contents_node.root[:length])
        else:
            chunk_count = (length + 31) // 32
            chunks = [contents_node.getter(to_gindex(i, contents_depth)) for i in range(chunk_count)]
            bytez = b"".join(ch.root for ch in chunks)[:length]
            return cls.decode_bytes(bytez)

    def get_backing(self) -> Node:
        return PairNode(
            subtree_fill_to_contents(pack_bytes_to_chunks(self), self.__class__.contents_depth()),
            uint256(len(self)).get_backing()
        )

    @classmethod
    def contents_depth(cls) -> int:
        raise NotImplementedError

    @classmethod
    def tree_depth(cls) -> int:
        return cls.contents_depth() + 1  # 1 extra for length mix-in

    @classmethod
    def default_node(cls) -> Node:
        return PairNode(zero_node(cls.contents_depth()), zero_node(0))  # mix-in 0 as list length

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        if key < 0 or key > cls.limit():
            raise KeyError
        return byte

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        depth = cls.tree_depth()
        byte_limit = cls.limit()
        if key < 0 or key >= byte_limit:
            raise KeyError
        chunk_i = key // 32
        return to_gindex(chunk_i, depth)

    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        return False

    @classmethod
    def min_byte_length(cls) -> int:
        return 0

    @classmethod
    def max_byte_length(cls) -> int:
        return cls.limit()

    @classmethod
    def deserialize(cls: Type[BL], stream: BinaryIO, scope: int) -> BL:
        return cls.decode_bytes(stream.read(scope))

    def value_byte_length(self) -> int:
        return len(self)



from remerkleable.tree import Gindex, Node, get_anchor_gindex, ROOT_GINDEX
from typing import List, Tuple, TypeVar

K = TypeVar('K')
History = List[Tuple[K, Node]]


def get_target_history(history: History, target: Gindex) -> History:
    """
    Fetch an ordered, keyed, history of nodes at the given target.
    The sequential equal values are reduced into a single value, and keyed by the first occurrence.
    This is done efficiently by first comparing the top of the tree,
     de-duplicating, and then recursively going deeper into the tree.
    :param history: An ordered keyed series of nodes, the root nodes to get the target from
    :param target: The target node to look up, pointing to the values to return.
    :return: the target values, keyed by the keys of the top nodes the target values first occurred in.
    """
    anchor = get_anchor_gindex(target)
    pivot = anchor >> 1
    unanchor = target ^ anchor
    direction_left = unanchor < pivot

    # Don't go deeper than the anchor. In this case we just return the (duplicate reduced) history.
    is_anchor = (anchor == ROOT_GINDEX)

    out: History = []
    last = None

    for key, node in history:
        if is_anchor:
            child_node = node
        else:
            if direction_left:
                child_node = node.get_left()
            else:
                child_node = node.get_right()
        if last is None or child_node.merkle_root() == last:
            continue
        out.append((key, child_node))
        last = child_node

    if not is_anchor:
        out = get_target_history(out, Gindex(pivot | unanchor))

    return out



# flake8:noqa F401  Ignore unused imports. Tests are a work in progress.

import pytest  # type: ignore

from random import Random

from remerkleable.complex import Container, Vector, List
from remerkleable.basic import boolean, bit, uint, byte, uint8, uint16, uint32, uint64, uint128, uint256,\
    OperationNotSupported
from remerkleable.bitfields import Bitvector, Bitlist
from remerkleable.byte_arrays import ByteVector, Bytes1, Bytes4, Bytes8, Bytes32, Bytes48, Bytes96
from remerkleable.core import BasicView, View
from remerkleable.tree import get_depth


def expect_op_error(fn, msg):
    try:
        fn()
        raise AssertionError(msg)
    except (ValueError, OperationNotSupported, AttributeError) as e:
        pass


def test_subclasses():
    for u in [uint8, uint16, uint32, uint64, uint128, uint256]:
        assert issubclass(u, int)
        assert issubclass(u, View)
        assert issubclass(u, BasicView)
    assert issubclass(boolean, BasicView)
    assert issubclass(boolean, View)

    for c in [Container, List, Vector, Bytes32]:
        assert issubclass(c, View)


def test_basic_instances():
    for u in [uint8, byte, uint16, uint32, uint64, uint128, uint256]:
        v = u(123)
        assert isinstance(v, int)
        assert isinstance(v, BasicView)
        assert isinstance(v, View)

    assert isinstance(boolean(True), BasicView)
    assert isinstance(boolean(False), BasicView)
    assert isinstance(bit(True), boolean)
    assert isinstance(bit(False), boolean)


def test_basic_value_bounds():
    max = {
        boolean: 2 ** 1,
        bit: 2 ** 1,
        uint8: 2 ** (8 * 1),
        byte: 2 ** (8 * 1),
        uint16: 2 ** (8 * 2),
        uint32: 2 ** (8 * 4),
        uint64: 2 ** (8 * 8),
        uint128: 2 ** (8 * 16),
        uint256: 2 ** (8 * 32),
    }
    for k, v in max.items():
        # this should work
        assert k(v - 1) == v - 1
        # but we do not allow overflows
        expect_op_error(lambda: k(v), f"no overflows allowed: type: {k}, value: {v}")

    for k, _ in max.items():
        # this should work
        assert k(0) == 0
        # but we do not allow underflows
        expect_op_error(lambda: k(-1), "no underflows allowed")

    for k, v in max.items():
        if v == 2:
            continue  # skip bool/bit
        half = v // 2
        # this should work
        assert k(half) + k(half-1) == v - 1
        # but 2x half == max, so this should fail
        expect_op_error(lambda: k(half) + k(half), f"no __add__ overflows allowed: type: {k}, value: {half}")
        # this should work
        assert k(half) - k(half) == 0
        assert k(half-1) - k(half-1) == 0
        # but underflow should not
        expect_op_error(lambda: k(half - 1) - k(half), f"no __sub__ underflows allowed: type: {k}, value: {half}")

    for k, v in max.items():
        if v == 2:
            continue  # skip bool/bit
        assert k(v-1) * k(1) == v-1
        assert k(v // 3) * 2 == (v // 3) * 2
        assert k((v // 2) - 1) * 2 == ((v // 2) - 1) * 2
        # but 2x max-1 should fail
        expect_op_error(lambda: k(v - 1) * 2, f"no __mul__ overflows allowed: type: {k}")
        # and 2x half too
        expect_op_error(lambda: k(v // 2) * 2, f"no __mul__ overflows allowed: type: {k}")
        expect_op_error(lambda: k(v // 2) // 0.5, f"no __floordiv__ with float overflows allowed: type: {k}")
        expect_op_error(lambda: k(v - 1) / 2.0, f"no __truediv__ allowed: type: {k}")


def test_container():
    class Foo(Container):
        a: uint8
        b: uint32

    empty = Foo()
    assert empty.a == uint8(0)
    assert empty.b == uint32(0)

    assert issubclass(Foo, Container)
    assert issubclass(Foo, View)

    assert Foo.is_fixed_byte_length()
    x = Foo(a=uint8(123), b=uint32(45))
    assert x.a == 123
    assert x.b == 45
    assert isinstance(x.a, uint8)
    assert isinstance(x.b, uint32)
    assert x.__class__.is_fixed_byte_length()

    class Bar(Container):
        a: uint8
        b: List[uint8, 1024]

    assert not Bar.is_fixed_byte_length()

    y = Bar(a=123, b=List[uint8, 1024](uint8(1), uint8(2)))
    assert y.a == 123
    assert isinstance(y.a, uint8)
    assert len(y.b) == 2
    assert isinstance(y.a, uint8)
    assert not y.__class__.is_fixed_byte_length()
    assert y.b[0] == 1
    v: List = y.b
    assert v.__class__.element_cls() == uint8
    assert v.__class__.limit() == 1024

    field_values = list(y)
    assert field_values == [y.a, y.b]

    f_a, f_b = y
    assert f_a == y.a
    assert f_b == y.b

    y.a = 42
    try:
        y.a = 256  # out of bounds
        assert False
    except ValueError:
        pass

    try:
        y.a = uint16(255)  # within bounds, wrong type
        assert False
    except ValueError:
        pass

    try:
        y.not_here = 5
        assert False
    except AttributeError:
        pass

    try:
        Foo(wrong_field_name=100)
        assert False
    except AttributeError:
        pass


def test_container_unpack():
    class Foo(Container):
        a: uint64
        b: uint8
        c: Vector[uint16, 123]

    foo = Foo(b=42)
    a, b, c = foo
    assert b == 42


def test_list():
    typ = List[uint64, 128]
    assert issubclass(typ, List)
    assert issubclass(typ, View)

    assert not typ.is_fixed_byte_length()

    assert len(typ()) == 0  # empty
    assert len(typ(uint64(0))) == 1  # single arg
    assert len(typ(uint64(i) for i in range(10))) == 10  # generator
    assert len(typ(uint64(0), uint64(1), uint64(2))) == 3  # args
    assert isinstance(typ(1, 2, 3, 4, 5)[4], uint64)  # coercion
    assert isinstance(typ(i for i in range(10))[9], uint64)  # coercion in generator

    v = typ(uint64(2), uint64(1))
    v[0] = uint64(123)
    assert v[0] == 123
    assert isinstance(v[0], uint64)

    assert isinstance(v, List)
    assert isinstance(v, View)

    assert len(typ([i for i in range(10)])) == 10  # cast py list to SSZ list

    foo = List[uint32, 128](0 for i in range(128))
    foo[0] = 123
    foo[1] = 654
    foo[127] = 222
    assert sum(foo) == 999
    try:
        foo[3] = 2 ** 32  # out of bounds
    except ValueError:
        pass

    for i in range(128):
        foo.pop()
        assert len(foo) == 128 - 1 - i
    for i in range(128):
        foo.append(uint32(i))
        assert len(foo) == i + 1
        assert foo[i] == i

    try:
        foo[3] = uint64(2 ** 32 - 1)  # within bounds, wrong type
        assert False
    except ValueError:
        pass

    try:
        foo[128] = 100
        assert False
    except IndexError:
        pass

    try:
        foo[-1] = 100  # valid in normal python lists
        assert False
    except IndexError:
        pass

    try:
        foo[128] = 100  # out of bounds
        assert False
    except IndexError:
        pass


def test_bytesn_subclass():
    class Root(Bytes32):
        pass

    # mypy does not like instance checks with parametrized generics, but they work, python can do everything
    assert isinstance(Root(b'\xab' * 32), Bytes32)  # type: ignore
    assert not isinstance(Root(b'\xab' * 32), Bytes48)  # type: ignore
    assert issubclass(Root(b'\xab' * 32).__class__, Bytes32)  # type: ignore
    assert issubclass(Root, Bytes32)  # type: ignore

    assert not issubclass(Bytes48, Bytes32)  # type: ignore

    assert len(Bytes32() + Bytes48()) == 80


def test_uint_math():
    assert uint8(0) + uint8(uint32(16)) == uint8(16)  # allow explicit casting to make invalid addition valid

    expect_op_error(lambda: uint8(0) - uint8(1), "no underflows allowed")
    expect_op_error(lambda: uint8(1) + uint8(255), "no overflows allowed")
    expect_op_error(lambda: uint8(0) + 256, "no overflows allowed")
    expect_op_error(lambda: uint8(42) + uint32(123), "no mixed types")
    expect_op_error(lambda: uint32(42) + uint8(123), "no mixed types")

    assert type(uint32(1234) + 56) == uint32


def test_container_depth():
    class SingleField(Container):
        foo: uint32

    assert SingleField.tree_depth() == 0

    class TwoField(Container):
        foo: uint32
        bar: uint64

    assert TwoField.tree_depth() == 1

    class ThreeField(Container):
        foo: uint32
        bar: uint64
        quix: uint8

    assert ThreeField.tree_depth() == 2

    class FourField(Container):
        foo: uint32
        bar: uint64
        quix: uint8
        more: uint32

    assert FourField.tree_depth() == 2

    class FiveField(Container):
        foo: uint32
        bar: uint64
        quix: uint8
        more: uint32
        fiv: uint8

    assert FiveField.tree_depth() == 3


@pytest.mark.parametrize("count, depth", [
    (0, 0), (1, 0), (2, 1), (3, 2), (4, 2), (5, 3), (6, 3), (7, 3), (8, 3), (9, 4)])
def test_tree_depth(count: int, depth: int):
    assert get_depth(count) == depth


def test_paths():
    class FourField(Container):
        foo: uint32
        bar: uint64
        quix: uint8
        more: uint32

    assert (FourField / 'foo').navigate_type() == uint32
    assert (FourField / 'bar').navigate_type() == uint64
    assert (FourField / 'foo').gindex() == 0b100
    assert (FourField / 'bar').gindex() == 0b101

    class Wrapper(Container):
        a: uint32
        b: FourField

    assert (Wrapper / 'a').navigate_type() == uint32
    assert (Wrapper / 'b').navigate_type() == FourField
    assert (Wrapper / 'b' / 'quix').navigate_type() == uint8

    assert (Wrapper / 'a').gindex() == 0b10
    assert (Wrapper / 'b' / 'more').gindex() == 0b1111

    assert ((Wrapper / 'b') / (FourField / 'quix')).gindex() == ((Wrapper / 'b') / 'quix').gindex()

    w = Wrapper(b=FourField(quix=42))
    assert (Wrapper / 'b' / 'quix').navigate_view(w) == 42

    assert (List[uint32, 123] / 0).navigate_type() == uint32
    try:
        (List[uint32, 123] / 123).navigate_type()
        assert False
    except KeyError:
        pass


def test_bitvector():
    for size in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
        for i in range(size):
            b = Bitvector[size]()
            b[i] = True
            assert bool(b[i]) == True, "set %d / %d" % (i, size)

        for i in range(size):
            b = Bitvector[size](True for i in range(size))
            b[i] = False
            assert bool(b[i]) == False, "unset %d / %d" % (i, size)


def test_bitvector_iter():
    rng = Random(123)
    for size in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
        # get a somewhat random bitvector
        bools = list(rng.randint(0, 1) == 1 for i in range(size))
        b = Bitvector[size](*bools)
        # Test iterator
        for i, bit in enumerate(b):
            assert bool(bit) == bools[i]


def test_bitlist():
    for size in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
        for i in range(size):
            b = Bitlist[size](False for i in range(size))
            b[i] = True
            assert bool(b[i]) == True, "set %d / %d" % (i, size)

        for i in range(size):
            b = Bitlist[size](True for i in range(size))
            b[i] = False
            assert bool(b[i]) == False, "unset %d / %d" % (i, size)


def test_bitlist_iter():
    rng = Random(123)
    for limit in [1, 2, 3, 4, 5, 6, 7, 8, 9, 15, 16, 17, 31, 32, 33, 63, 64, 65, 511, 512, 513, 1023, 1024, 1025]:
        for length in [0, 1, limit // 2, limit]:
            # get a somewhat random bitvector
            bools = list(rng.randint(0, 1) == 1 for i in range(length))
            b = Bitlist[limit](*bools)
            # Test iterator
            for i, bit in enumerate(b):
                assert bool(bit) == bools[i]


def test_container_inheritance():
    class Foo(Container):
        a: uint64
        b: uint32

    class Bar(Foo):
        c: uint8

    assert Foo.fields() == {'a': uint64, 'b': uint32}
    assert Bar.fields() == {'a': uint64, 'b': uint32, 'c': uint8}
    assert Foo._field_indices == {'a': 0, 'b': 1}
    assert Bar._field_indices == {'a': 0, 'b': 1, 'c': 2}

    foo = Foo(a=0xaabbccdd11223344, b=0x55667788)
    assert foo.encode_bytes().hex() == "44332211ddccbbaa88776655"  # little endian!
    bar = Bar(c=0x99)
    assert bar.encode_bytes().hex() == "00000000000000000000000099"  # inits missing fields still
    bar2 = Bar(a=0xaabbccdd11223344, b=0x55667788, c=0x99)
    assert bar2.encode_bytes().hex() == "44332211ddccbbaa8877665599"

    class ChocoBar(Bar):
        aa: uint64

    assert ChocoBar.fields() == {'a': uint64, 'b': uint32, 'c': uint8, 'aa': uint64}
    assert ChocoBar._field_indices == {'a': 0, 'b': 1, 'c': 2, 'aa': 3}
    cb = ChocoBar(a=0xaabbccdd11223344, b=0x55667788, c=0xaf, aa=0x0102030405060708)
    assert cb.encode_bytes().hex() == "44332211ddccbbaa88776655af0807060504030201"

    # multiple inheritance

    class Mixin(Foo):
        m: uint8

    foo_mix = Mixin(a=0xaabbccdd11223344, b=0x55667788, m=0x99)
    assert foo_mix.encode_bytes().hex() == "44332211ddccbbaa8877665599"

    class Combined(Bar, Mixin):
        ...

    assert Combined.fields() == {'a': uint64, 'b': uint32, 'c': uint8, 'm': uint8}
    assert Combined._field_indices == {'a': 0, 'b': 1, 'c': 2, 'm': 3}

    combi = Combined(a=0xaabbccdd11223344, b=0x55667788, c=0x99, m=0x42)
    assert combi.encode_bytes().hex() == "44332211ddccbbaa887766559942"

    class Switcheroo(Mixin, Bar):
        ...

    assert Switcheroo.fields() == {'a': uint64, 'b': uint32, 'm': uint8, 'c': uint8}
    assert Switcheroo._field_indices == {'a': 0, 'b': 1, 'm': 2, 'c': 3}

    combi_switch = Switcheroo(a=0xaabbccdd11223344, b=0x55667788, c=0x99, m=0x42)
    assert combi_switch.encode_bytes().hex() == "44332211ddccbbaa887766554299"
    # fields are switched, while being different, hash tree roots should be different then
    assert combi.hash_tree_root() != combi_switch.hash_tree_root()

    class Duplicates(ChocoBar, Mixin, Foo):
        ...

    assert Duplicates.fields() == {'a': uint64, 'b': uint32, 'c': uint8, 'aa': uint64, 'm': uint8}
    assert Duplicates._field_indices == {'a': 0, 'b': 1, 'c': 2, 'aa': 3, 'm': 4}

    # overriding with different type
    class FancyChocoBar(ChocoBar):
        aa: uint128  # type: ignore

    assert FancyChocoBar.fields() == {'a': uint64, 'b': uint32, 'c': uint8, 'aa': uint128}
    assert FancyChocoBar._field_indices == {'a': 0, 'b': 1, 'c': 2, 'aa': 3}

    fcb = FancyChocoBar(a=0xaabbccdd11223344, b=0x55667788, c=0xaf, aa=0x0102030405060708a1a2a3a4a5a6a7a8)
    assert fcb.encode_bytes().hex() == "44332211ddccbbaa88776655afa8a7a6a5a4a3a2a10807060504030201"

    # overriding and extending
    class ExtendedFancyChocoBar(ChocoBar):
        more: uint16
        aa: uint128  # type: ignore

    # overriden field must stay in place
    assert ExtendedFancyChocoBar.fields() == {'a': uint64, 'b': uint32, 'c': uint8, 'aa': uint128, 'more': uint16}
    assert ExtendedFancyChocoBar._field_indices == {'a': 0, 'b': 1, 'c': 2, 'aa': 3, 'more': 4}

    efcb = ExtendedFancyChocoBar(a=0xaabbccdd11223344, b=0x55667788, c=0xaf,
                                 aa=0x0102030405060708a1a2a3a4a5a6a7a8, more=0xe1e2)
    assert efcb.encode_bytes().hex() == "44332211ddccbbaa88776655afa8a7a6a5a4a3a2a10807060504030201e2e1"



from typing import cast, BinaryIO, List as PyList, Any, TypeVar, Type
from types import GeneratorType
from collections.abc import Sequence as ColSequence
import io
from remerkleable.core import BackedView, FixedByteLengthViewHelper, \
    pack_bits_to_chunks, View, ObjType, ObjParseException
from remerkleable.tree import Node, PairNode, zero_node, Gindex, to_gindex, Link, RootNode, NavigationError,\
    Root, subtree_fill_to_contents, subtree_fill_to_length, get_depth
from remerkleable.basic import boolean, uint256
from remerkleable.readonly_iters import BitfieldIter

V = TypeVar('V', bound=View)


def _new_chunk_with_bit(chunk: Node, i: int, v: boolean) -> Node:
    new_chunk_root = bytearray(bytes(chunk.root))  # mutable copy
    if v:
        new_chunk_root[(i & 0xff) >> 3] |= 1 << (i & 0x7)
    else:
        new_chunk_root[(i & 0xff) >> 3] &= (~(1 << (i & 0x7))) & 0xff
    return RootNode(Root(new_chunk_root))


# alike to the SubtreeView, but specialized to work on individual bits of chunks, instead of complex/basic types.
class BitsView(BackedView, ColSequence):

    @classmethod
    def coerce_view(cls: Type[V], v: Any) -> V:
        return cls(*v)

    @classmethod
    def tree_depth(cls) -> int:
        raise NotImplementedError

    def length(self) -> int:
        raise NotImplementedError

    def get(self, i: int) -> boolean:
        ll = self.length()
        if i >= ll:
            raise NavigationError(f"cannot get bit {i} in bits of length {ll}")
        chunk_i = i >> 8
        chunk = self.get_backing().getter(to_gindex(chunk_i, self.__class__.tree_depth()))
        chunk_byte = chunk.root[(i & 0xff) >> 3]
        return boolean((chunk_byte >> (i & 0x7)) & 1)

    def set(self, i: int, v: boolean) -> None:
        ll = self.length()
        if i >= ll:
            raise NavigationError(f"cannot set bit {i} in bits of length {ll}")
        chunk_i = i >> 8
        chunk_setter_link: Link = self.get_backing().setter(to_gindex(chunk_i, self.__class__.tree_depth()))
        chunk = self.get_backing().getter(to_gindex(chunk_i, self.__class__.tree_depth()))
        new_chunk = _new_chunk_with_bit(chunk, i & 0xff, v)
        self.set_backing(chunk_setter_link(new_chunk))

    def __len__(self):
        return self.length()

    def __getitem__(self, k):
        length = self.length()
        if isinstance(k, slice):
            start = 0 if k.start is None else k.start
            if start < 0:
                start = start % length
            end = length if k.stop is None else k.stop
            if end < 0:
                end = end % length
            return [self.get(i) for i in range(start, end)]
        else:
            return self.get(k)

    def __setitem__(self, k, v):
        length = self.length()
        if type(k) == slice:
            i = 0 if k.start is None else k.start
            end = length if k.stop is None else k.stop
            for item in v:
                self.set(i, item)
                i += 1
            if i != end:
                raise Exception("failed to do full slice-set, not enough values")
        else:
            self.set(k, v)

    def encode_bytes(self) -> bytes:
        stream = io.BytesIO()
        self.serialize(stream)
        stream.seek(0)
        return stream.read()

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        stream = io.BytesIO()
        stream.write(bytez)
        stream.seek(0)
        return cls.deserialize(stream, len(bytez))

    @classmethod
    def from_obj(cls: Type[V], obj: ObjType) -> V:
        if not isinstance(obj, (list, tuple, str)):
            raise ObjParseException(f"obj '{obj}' is not a list, tuple or str")
        if isinstance(obj, str):
            if obj.startswith('0x'):
                return cls.decode_bytes(bytes.fromhex(obj[2:]))
            obj = [c == '1' for c in obj]
        return cls(obj)  # type: ignore

    def to_obj(self) -> ObjType:
        return '0x' + self.encode_bytes().hex()

    def navigate_view(self, key: Any) -> View:
        return boolean(self.__getitem__(key))


class Bitlist(BitsView):
    def __new__(cls, *args, **kwargs):
        vals = list(args)
        if len(vals) > 0:
            if len(vals) == 1 and isinstance(vals[0], (GeneratorType, list, tuple)):
                vals = list(vals[0])
            limit = cls.limit()
            if len(vals) > limit:
                raise Exception(f"too many bitlist inputs: {len(vals)}, limit is: {limit}")
            input_bits = list(map(bool, vals))
            input_nodes = pack_bits_to_chunks(input_bits)
            contents = subtree_fill_to_contents(input_nodes, cls.contents_depth())
            kwargs['backing'] = PairNode(contents, uint256(len(input_bits)).get_backing())
        return super().__new__(cls, **kwargs)

    def __class_getitem__(cls, limit) -> Type["Bitlist"]:
        class SpecialBitlistView(Bitlist):
            @classmethod
            def limit(cls) -> int:
                return limit

        return SpecialBitlistView

    def __iter__(self):
        return BitfieldIter(self.get_backing().get_left(), self.contents_depth(), self.length())

    @classmethod
    def contents_depth(cls) -> int:  # depth excluding the length mix-in
        return get_depth((cls.limit() + 255) // 256)

    @classmethod
    def tree_depth(cls) -> int:
        return cls.contents_depth() + 1  # 1 extra for length mix-in

    @classmethod
    def limit(cls) -> int:
        raise NotImplementedError

    @classmethod
    def default_node(cls) -> Node:
        return PairNode(zero_node(cls.contents_depth()), zero_node(0))  # mix-in 0 as list length

    @classmethod
    def type_repr(cls) -> str:
        return f"Bitlist[{cls.limit()}]"

    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        return False

    @classmethod
    def min_byte_length(cls) -> int:
        return 1  # the delimiting bit will always require at least 1 byte

    @classmethod
    def max_byte_length(cls) -> int:
        # maximum bit count in bytes rounded up + delimiting bit
        return (cls.limit() + 7 + 1) // 8

    def length(self) -> int:
        ll_node = super().get_backing().get_right()
        ll = cast(uint256, uint256.view_from_backing(node=ll_node, hook=None))
        return int(ll)

    def append(self, v: boolean):
        ll = self.length()
        if ll >= self.__class__.limit():
            raise Exception("list is maximum capacity, cannot append")
        i = ll
        chunk_i = i // 256
        target: Gindex = to_gindex(chunk_i, self.__class__.tree_depth())
        if i & 0xff == 0:
            set_last = self.get_backing().setter(target, expand=True)
            next_backing = set_last(_new_chunk_with_bit(zero_node(0), 0, v))
        else:
            set_last = self.get_backing().setter(target)
            chunk = self.get_backing().getter(target)
            next_backing = set_last(_new_chunk_with_bit(chunk, i & 0xff, v))
        set_length = next_backing.rebind_right
        new_length = uint256(ll + 1).get_backing()
        next_backing = set_length(new_length)
        self.set_backing(next_backing)

    def pop(self):
        ll = self.length()
        if ll == 0:
            raise Exception("list is empty, cannot pop")
        i = ll - 1
        chunk_i = i // 256
        target: Gindex = to_gindex(chunk_i, self.__class__.tree_depth())
        if i & 0xff == 0:
            set_last = self.get_backing().setter(target)
            next_backing = set_last(zero_node(0))
        else:
            set_last = self.get_backing().setter(target)
            chunk = self.get_backing().getter(target)
            next_backing = set_last(_new_chunk_with_bit(chunk, ll & 0xff, boolean(False)))

        # if possible, summarize
        can_summarize = (target & 1) == 0
        if can_summarize:
            # summarize to the highest node possible.
            # I.e. the resulting target must be a right-hand, unless it's the only content node.
            while (target & 1) == 0 and target != 0b10:
                target >>= 1
            summary_fn = next_backing.summarize_into(target)
            next_backing = summary_fn()

        set_length = next_backing.rebind_right
        new_length = uint256(ll - 1).get_backing()
        next_backing = set_length(new_length)
        self.set_backing(next_backing)

    def get(self, i: int) -> boolean:
        if i < 0 or i >= self.length():
            raise IndexError
        try:
            return super().get(i)
        except NavigationError:
            raise IndexError

    def set(self, i: int, v: boolean) -> None:
        if i < 0 or i >= self.length():
            raise IndexError
        try:
            super().set(i, v)
        except NavigationError:
            raise IndexError

    def __repr__(self):
        try:
            length = self.length()
        except NavigationError:
            return f"Bitlist[{self.__class__.limit()}]~partial"
        try:
            bitstr = ''.join('1' if self.get(i) else '0' for i in range(length))
        except NavigationError:
            bitstr = " *partial bits* "
        return f"Bitlist[{self.__class__.limit()}]({length} bits: {bitstr})"

    def value_byte_length(self) -> int:
        # bit count in bytes rounded up + delimiting bit
        return (self.length() + 7 + 1) // 8

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        stream = io.BytesIO()
        stream.write(bytez)
        stream.seek(0)
        return cls.deserialize(stream, len(bytez))

    @classmethod
    def deserialize(cls: Type[V], stream: BinaryIO, scope: int) -> V:
        if scope < 1:
            raise Exception("cannot have empty scope for bitlist, need at least a delimiting bit")
        if scope > cls.max_byte_length():
            raise Exception(f"scope is too large: {scope}, max bitlist byte length is: {cls.max_byte_length()}")
        chunks: PyList[Node] = []
        bytelen = scope - 1  # excluding the last byte (which contains the delimiting bit)
        while scope > 32:
            chunks.append(RootNode(Root(stream.read(32))))
            scope -= 32
        # scope is [1, 32] here
        last_chunk_part = stream.read(scope)
        last_byte = int(last_chunk_part[scope-1])
        if last_byte == 0:
            raise Exception("last byte must not be 0: bitlist requires delimiting bit")
        last_byte_bitlen = last_byte.bit_length() - 1  # excluding the delimiting bit
        bitlen = bytelen * 8 + last_byte_bitlen
        if bitlen % 256 != 0:
            last_chunk = last_chunk_part[:scope-1] +\
                         (last_byte ^ (1 << last_byte_bitlen)).to_bytes(length=1, byteorder='little')
            last_chunk += b"\x00" * (32 - len(last_chunk))
            chunks.append(RootNode(Root(last_chunk)))
        if bitlen > cls.limit():
            raise Exception(f"bitlist too long: {bitlen}, delimiting bit is over limit ({cls.limit()})")
        contents = subtree_fill_to_contents(chunks, cls.contents_depth())
        backing = PairNode(contents, uint256(bitlen).get_backing())
        return cls.view_from_backing(backing)

    def serialize(self, stream: BinaryIO) -> int:
        backing = self.get_backing()
        bitlen = self.length()
        chunk_count = (bitlen + 255) // 256  # excludes delimit bit, this is the backing, not the serialized form
        byte_len = (bitlen + 7) // 8
        tree_depth = self.tree_depth()
        full_chunks_count = max(0, chunk_count - 1)
        for chunk_index in range(full_chunks_count):
            chunk = backing.getter(to_gindex(chunk_index, tree_depth))
            stream.write(chunk.root)
        if chunk_count > 0:
            last_chunk = backing.getter(to_gindex(chunk_count - 1, tree_depth))
            # write the last chunk, may not be a full chunk
            last_chunk_bytes_count = byte_len - (full_chunks_count * 32)
            bytez = last_chunk.root[:last_chunk_bytes_count]
            # add in delimiting bit
            if bitlen % 8 == 0:
                bytez += b"\x01"
            else:
                bytez = bytez[:len(bytez) - 1] +\
                        (bytez[len(bytez) - 1] ^ (1 << (bitlen % 8))).to_bytes(length=1, byteorder='little')
            stream.write(bytez)
        else:
            stream.write(b"\x01")  # empty bitlist still has a delimiting bit
        return (bitlen + 7 + 1) // 8  # includes delimit bit in length computation

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        bit_limit = cls.limit()
        if key < 0 or key >= bit_limit:
            raise KeyError
        return boolean

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        depth = cls.tree_depth()
        bit_limit = cls.limit()
        if key < 0 or key >= bit_limit:
            raise KeyError
        chunk_i = key // 256
        return to_gindex(chunk_i, depth)


class Bitvector(BitsView, FixedByteLengthViewHelper):
    def __new__(cls, *args, **kwargs):
        vals = list(args)
        if len(vals) > 0:
            if len(vals) == 1 and isinstance(vals[0], (GeneratorType, list, tuple)):
                vals = list(vals[0])
            veclen = cls.vector_length()
            if len(vals) != veclen:
                raise Exception(f"incorrect bitvector input: {len(vals)} bits, vector length is: {veclen}")
            input_bits = list(map(bool, vals))
            input_nodes = pack_bits_to_chunks(input_bits)
            kwargs['backing'] = subtree_fill_to_contents(input_nodes, cls.tree_depth())
        return super().__new__(cls, **kwargs)

    def __class_getitem__(cls, length) -> Type["Bitvector"]:
        if length <= 0:
            raise Exception(f"invalid bitvector length: {length}")

        class SpecialBitvectorView(Bitvector):
            @classmethod
            def vector_length(cls) -> int:
                return length

        return SpecialBitvectorView

    def __iter__(self):
        return BitfieldIter(self.get_backing(), self.__class__.tree_depth(), self.__class__.vector_length())

    @classmethod
    def tree_depth(cls) -> int:
        return get_depth((cls.vector_length() + 255) // 256)

    @classmethod
    def vector_length(cls) -> int:
        raise NotImplementedError

    @classmethod
    def default_node(cls) -> Node:
        return subtree_fill_to_length(zero_node(0), cls.tree_depth(), ((cls.vector_length() + 255) // 256))

    @classmethod
    def type_repr(cls) -> str:
        return f"Bitvector[{cls.vector_length()}]"

    @classmethod
    def type_byte_length(cls) -> int:
        return (cls.vector_length() + 7) // 8

    def length(self) -> int:
        return self.__class__.vector_length()

    def get(self, i: int) -> boolean:
        if i < 0 or i >= self.length():
            raise IndexError
        return super().get(i)

    def set(self, i: int, v: boolean) -> None:
        if i < 0 or i >= self.length():
            raise IndexError
        super().set(i, v)

    def __repr__(self):
        length = self.length()
        try:
            bitstr = ''.join('1' if self.get(i) else '0' for i in range(length))
        except NavigationError:
            bitstr = " *partial bits* "
        return f"Bitvector[{length}]({bitstr})"

    @classmethod
    def deserialize(cls: Type[V], stream: BinaryIO, scope: int) -> V:
        if scope != cls.type_byte_length():
            raise Exception(f"scope is invalid: {scope}, bitvector byte length is: {cls.type_byte_length()}")
        chunks: PyList[Node] = []
        bytelen = scope - 1  # excluding the last byte
        while scope > 32:
            chunks.append(RootNode(Root(stream.read(32))))
            scope -= 32
        # scope is [1, 32] here
        last_chunk_part = stream.read(scope)
        last_byte = int(last_chunk_part[scope-1])
        bitlen = bytelen * 8 + last_byte.bit_length()
        if bitlen > cls.vector_length():
            raise Exception(f"bitvector too long: {bitlen}, last byte has bits over bit length ({cls.vector_length()})")
        last_chunk = last_chunk_part + (b"\x00" * (32 - len(last_chunk_part)))
        chunks.append(RootNode(Root(last_chunk)))
        backing = subtree_fill_to_contents(chunks, cls.tree_depth())
        return cls.view_from_backing(backing)

    def serialize(self, stream: BinaryIO) -> int:
        backing = self.get_backing()
        bitlen = self.length()
        chunk_count = (bitlen + 255) // 256  # excludes delimit bit, this is the backing, not the serialized form
        byte_len = (bitlen + 7) // 8
        tree_depth = self.tree_depth()
        full_chunks_count = max(0, chunk_count - 1)
        for chunk_index in range(full_chunks_count):
            chunk: Node = backing.getter(to_gindex(chunk_index, tree_depth))
            stream.write(chunk.root)
        if chunk_count > 0:
            last_chunk = backing.getter(to_gindex(chunk_count - 1, tree_depth))
            # write the last chunk, may not be a full chunk
            last_chunk_bytes_count = byte_len - (full_chunks_count * 32)
            stream.write(last_chunk.root[:last_chunk_bytes_count])
        return byte_len

    @classmethod
    def navigate_type(cls, key: Any) -> Type[View]:
        bit_limit = cls.vector_length()
        if key < 0 or key >= bit_limit:
            raise KeyError
        return boolean

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        depth = cls.tree_depth()
        bit_len = cls.vector_length()
        if key < 0 or key >= bit_len:
            raise KeyError
        chunk_i = key // 256
        return to_gindex(chunk_i, depth)



from setuptools import setup, find_packages

with open("README.rst", "rt", encoding="utf8") as f:
    readme = f.read()

setup(
    name="remerkleable",
    description="Typed mutable SSZ views over cached and immutable binary merkle trees",
    version="0.1.18",
    long_description=readme,
    long_description_content_type="text/x-rst",
    author="protolambda",
    author_email="proto+pip@protolambda.com",
    url="https://github.com/protolambda/remerkleable",
    python_requires=">=3.8, <4",
    license="MIT",
    packages=find_packages(),
    py_modules=["remerkleable"],
    tests_require=[],
    extras_require={
        "testing": ["pytest"],
        "linting": ["flake8", "mypy"],
        "docs": ["sphinx", "sphinx-autodoc-typehints", "pallets_sphinx_themes", "sphinx_issues"]
    },
    install_requires=[],
    include_package_data=True,
    keywords=["merkle", "merkleize", "merkle-tree", "merkle-trie", "trie", "ssz", "hash-tree-root", "eth2"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
    ],
)



from typing import Callable, Optional, Any, cast, List as PyList, BinaryIO,\
    TypeVar, Type, Protocol, runtime_checkable, Union

# noinspection PyUnresolvedReferences
from typing import _ProtocolMeta  # type: ignore

from remerkleable.tree import Node, Root, RootNode, zero_node, concat_gindices, Gindex
from itertools import zip_longest
from typing import Iterable, Tuple

OFFSET_BYTE_LENGTH = 4


V = TypeVar('V', bound="View")


class ViewMeta(_ProtocolMeta):
    def __truediv__(self, other) -> "Path":
        return Path.from_raw_path(anchor=cast(Type[View], self), path=[other])


ObjType = Union[dict, list, tuple, str, int, bool, None]


class ObjParseException(Exception):
    pass


class Path(object):
    anchor: Type["View"]
    path: PyList[Tuple[Any, Type["View"]]]  # (key, type) tuples.

    def __init__(self, anchor: Type["View"], path: Optional[PyList[Tuple[Any, Type["View"]]]] = None):
        self.anchor = anchor
        if path is None:
            self.path = []
        else:
            self.path = path

    @staticmethod
    def from_raw_path(anchor: Type["View"], path: PyList[Any]) -> "Path":
        parsed_path = []
        t = anchor
        for step in path:
            t = t.navigate_type(step)
            parsed_path.append((step, t))
        return Path(anchor=anchor, path=parsed_path)

    def __truediv__(self, other) -> "Path":
        if isinstance(other, Path):
            return Path(anchor=self.anchor, path=self.path + other.path)
        else:
            last = self.anchor if len(self.path) == 0 else self.path[-1][1]
            return Path(anchor=self.anchor, path=self.path + [(other, last.navigate_type(other))])

    def gindex(self, view: Optional["View"] = None) -> Gindex:
        step_gindices = []
        if view is None:
            type_end: Type[View] = self.anchor
            for step, typ in self.path:
                gindex = type_end.key_to_static_gindex(step)
                step_gindices.append(gindex)
                type_end = typ
        else:
            view_end: View = view
            for step, _ in self.path:
                gindex = view_end.key_to_dynamic_gindex(step)
                step_gindices.append(gindex)
                view_end = view_end.navigate_view(step)
        return concat_gindices(step_gindices)

    def navigate_type(self) -> Type["View"]:
        return self.path[-1][1]

    def navigate_view(self, v: "View") -> "View":
        for step, _ in self.path:
            v = v.navigate_view(step)
        return v


HV = TypeVar('HV', bound="View")
ViewHook = Callable[[HV], None]


@runtime_checkable
class View(Protocol, metaclass=ViewMeta):

    @classmethod
    def coerce_view(cls: Type[V], v: Any) -> V:
        ...

    @classmethod
    def default_node(cls) -> Node:
        ...

    @classmethod
    def view_from_backing(cls: Type[V], node: Node, hook: Optional["ViewHook[V]"] = None) -> V:
        ...

    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        ...

    @classmethod
    def min_byte_length(cls) -> int:
        ...

    @classmethod
    def max_byte_length(cls) -> int:
        ...

    @classmethod
    def decode_bytes(cls: Type[V], bytez: bytes) -> V:
        ...

    @classmethod
    def deserialize(cls: Type[V], stream: BinaryIO, scope: int) -> V:
        ...

    @classmethod
    def from_obj(cls: Type[V], obj: ObjType) -> V:
        ...

    @classmethod
    def type_repr(cls) -> str:
        ...

    @classmethod
    def navigate_type(cls, key: Any) -> Type["View"]:
        raise Exception(f"cannot type-navigate into a {cls.type_repr()}, key: '{key}'")

    @classmethod
    def key_to_static_gindex(cls, key: Any) -> Gindex:
        raise Exception(f"cannot get static gindex into type {cls.type_repr()}, key: '{key}'")

    @classmethod
    def default(cls: Type[V], hook: Optional[ViewHook[V]]) -> V:
        return cls.view_from_backing(cls.default_node(), hook)

    def get_backing(self) -> Node:
        raise NotImplementedError

    def set_backing(self, value):
        raise NotImplementedError

    def copy(self: V) -> V:
        return self.__class__.view_from_backing(self.get_backing())

    @classmethod
    def type_byte_length(cls) -> int:
        raise Exception("type is dynamic length, or misses overrides. Cannot get type byte length.")

    def value_byte_length(self) -> int:
        raise NotImplementedError

    def __bytes__(self):
        return self.encode_bytes()

    def encode_bytes(self) -> bytes:
        raise NotImplementedError

    def serialize(self, stream: BinaryIO) -> int:
        out = self.encode_bytes()
        stream.write(out)
        return len(out)

    def to_obj(self) -> ObjType:
        raise NotImplementedError

    def navigate_view(self, key: Any) -> "View":
        raise Exception(f"cannot view-navigate into {self}, key: '{key}'")

    def key_to_dynamic_gindex(self, key: Any) -> Gindex:
        return self.__class__.key_to_static_gindex(key)

    def hash_tree_root(self) -> Root:
        return self.get_backing().merkle_root()

    def __eq__(self, other):
        # TODO: should we check types here?
        if not isinstance(other, View):
            other = self.__class__.coerce_view(other)
        return self.hash_tree_root() == other.hash_tree_root()

    def __hash__(self):
        return hash(self.hash_tree_root())


class FixedByteLengthViewHelper(View, Protocol):
    @classmethod
    def is_fixed_byte_length(cls) -> bool:
        return True

    @classmethod
    def min_byte_length(cls) -> int:
        return cls.type_byte_length()

    @classmethod
    def max_byte_length(cls) -> int:
        return cls.type_byte_length()

    @classmethod
    def deserialize(cls: Type[V], stream: BinaryIO, scope: int) -> V:
        n = cls.type_byte_length()
        if n != scope:
            raise Exception(f"scope {scope} is not valid for expected byte length {n}")
        return cls.decode_bytes(stream.read(n))

    def value_byte_length(self) -> int:
        return self.type_byte_length()


BackedV = TypeVar('BackedV', bound="BackedView")


class BackedView(View):
    _hook: Optional[ViewHook]
    _backing: Node

    @classmethod
    def view_from_backing(cls: Type[BackedV], node: Node, hook: Optional[ViewHook] = None) -> BackedV:
        return cls(backing=node, hook=hook)

    def __new__(cls, backing: Optional[Node] = None, hook: Optional[ViewHook] = None, **kwargs):
        if backing is None:
            backing = cls.default_node()
        out = super().__new__(cls, **kwargs)  # type: ignore
        out._backing = backing
        out._hook = hook
        return out

    def get_backing(self) -> Node:
        return self._backing

    def set_backing(self, value):
        self._backing = value
        # Propagate up the change if the view is hooked to a super view
        if self._hook is not None:
            self._hook(self)


BV = TypeVar('BV', bound="BasicView")


@runtime_checkable
class BasicView(FixedByteLengthViewHelper, Protocol):

    @classmethod
    def default_node(cls) -> Node:
        return zero_node(0)

    @classmethod
    def view_from_backing(cls: Type[BV], node: Node, hook: Optional[ViewHook[BV]] = None) -> BV:
        size = cls.type_byte_length()
        return cls.decode_bytes(node.root[0:size])

    @classmethod
    def basic_view_from_backing(cls: Type[BV], node: Node, i: int) -> BV:
        size = cls.type_byte_length()
        return cls.decode_bytes(node.root[i*size:(i+1)*size])

    @classmethod
    def pack_views(cls: Type[BV], views: PyList[BV]) -> PyList[Node]:
        return list(pack_ints_to_chunks((cast(int, v) for v in views), 32 // cls.type_byte_length()))

    def copy(self: V) -> V:
        return self  # basic views do not have to be copied, they are immutable

    def backing_from_base(self, base: Node, i: int) -> Node:
        section_bytez = self.encode_bytes()
        chunk_bytez = base.root[:len(section_bytez)*i] + section_bytez + base.root[len(section_bytez)*(i+1):]
        return RootNode(Root(chunk_bytez))

    def get_backing(self) -> Node:
        bytez = self.encode_bytes()
        return RootNode(Root(bytez + b"\x00" * (32 - len(bytez))))

    def set_backing(self, value):
        raise Exception("cannot change the backing of a basic view")


# recipe from more-itertools, should have been in itertools really.
def grouper(items: Iterable, n: int, fillvalue=None) -> Iterable[Tuple]:
    """Collect data into fixed-length chunks or blocks
       grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"""
    args = [iter(items)] * n
    # The *same* iterator is referenced n times, thus zip produces tuples of n elements from the same iterator
    return zip_longest(*args, fillvalue=fillvalue)


def pack_ints_to_chunks(items: Iterable[int], items_per_chunk: int) -> PyList[Node]:
    item_byte_len = 32 // items_per_chunk
    return [RootNode(Root(b"".join(v.to_bytes(length=item_byte_len, byteorder='little') for v in chunk_elems)))
            for chunk_elems in grouper(items, items_per_chunk, fillvalue=0)]


def bits_to_byte_int(byte: Tuple[bool, bool, bool, bool, bool, bool, bool, bool]) -> int:
    return sum([byte[i] << i for i in range(0, 8)])


def byte_int_to_byte(b: int) -> bytes:
    return b.to_bytes(length=1, byteorder='little')


def pack_bits_to_chunks(items: Iterable[bool]) -> PyList[Node]:
    # grouper returns tuples of N=8, bits_to_byte_int takes tuples of 8, but mypy does not follow.
    return pack_byte_ints_to_chunks(map(bits_to_byte_int, grouper(items, 8, fillvalue=0)))  # type: ignore


def pack_byte_ints_to_chunks(items: Iterable[int]) -> PyList[Node]:
    return [RootNode(Root(b"".join(map(byte_int_to_byte, chunk_bytes))))
            for chunk_bytes in grouper(items, 32, fillvalue=0)]


def pack_bytes_to_chunks(bytez: bytes) -> PyList[Node]:
    full_chunks_byte_len = (len(bytez) >> 5) << 5
    out: PyList[Node] = [RootNode(Root(bytez[i:i+32])) for i in range(0, full_chunks_byte_len, 32)]
    if len(bytez) != full_chunks_byte_len:
        out.append(RootNode(Root(bytez[full_chunks_byte_len:] + (b"\x00" * (32 - (len(bytez) - full_chunks_byte_len))))))
    return out



from typing import Type, cast
from remerkleable.core import View, BackedView, BasicView
from remerkleable.tree import Link, to_gindex


class SubtreeView(BackedView):
    @classmethod
    def is_packed(cls) -> bool:
        raise NotImplementedError

    @classmethod
    def tree_depth(cls) -> int:
        raise NotImplementedError

    @classmethod
    def item_elem_cls(cls, i: int) -> Type[View]:
        raise NotImplementedError

    def get(self, i: int) -> View:
        elem_type: Type[View] = self.item_elem_cls(i)
        # basic types are more complicated: we operate on subsections packed into a bottom chunk
        if self.is_packed():
            elems_per_chunk = 32 // elem_type.type_byte_length()
            chunk_i = i // elems_per_chunk
            chunk = self.get_backing().getter(to_gindex(chunk_i, self.tree_depth()))
            return cast(Type[BasicView], elem_type).basic_view_from_backing(chunk, i % elems_per_chunk)
        else:
            return elem_type.view_from_backing(
                self.get_backing().getter(to_gindex(i, self.tree_depth())), lambda v: self.set(i, v))

    def set(self, i: int, v: View) -> None:
        elem_type: Type[View] = self.item_elem_cls(i)
        # if not the right type, try to coerce it
        if not isinstance(v, elem_type):
            v = elem_type.coerce_view(v)
        if self.is_packed():
            # basic types are more complicated: we operate on a subsection of a bottom chunk
            if isinstance(v, BasicView):
                elems_per_chunk = 32 // v.type_byte_length()
                chunk_i = i // elems_per_chunk
                target = to_gindex(chunk_i, self.tree_depth())
                chunk_setter_link: Link = self.get_backing().setter(target)
                chunk = self.get_backing().getter(target)
                new_chunk = v.backing_from_base(chunk, i % elems_per_chunk)
                self.set_backing(chunk_setter_link(new_chunk))
            else:
                raise Exception("cannot pack subtree elements that are not basic types")
        else:
            setter_link: Link = self.get_backing().setter(to_gindex(i, self.tree_depth()))
            self.set_backing(setter_link(v.get_backing()))
